# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.protobuf import timestamp_pb2  # type: ignore


__protobuf__ = proto.module(
    package='google.events.cloud.notebooks.v1',
    manifest={
        'Environment',
        'VmImage',
        'ContainerImage',
        'Runtime',
        'RuntimeAcceleratorConfig',
        'EncryptionConfig',
        'LocalDisk',
        'LocalDiskInitializeParams',
        'RuntimeAccessConfig',
        'RuntimeSoftwareConfig',
        'RuntimeMetrics',
        'RuntimeShieldedInstanceConfig',
        'VirtualMachine',
        'VirtualMachineConfig',
        'ExecutionTemplate',
        'Execution',
        'ReservationAffinity',
        'Instance',
        'Schedule',
        'RuntimeEventData',
        'ExecutionEventData',
        'InstanceEventData',
        'ScheduleEventData',
        'EnvironmentEventData',
    },
)


class Environment(proto.Message):
    r"""Definition of a software environment that is used to start a
    notebook instance.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        name (str):
            Output only. Name of this environment. Format:
            ``projects/{project_id}/locations/{location}/environments/{environment_id}``
        display_name (str):
            Display name of this environment for the UI.
        description (str):
            A brief description of this environment.
        vm_image (google.events.cloud.notebooks_v1.types.VmImage):
            Use a Compute Engine VM image to start the
            notebook instance.

            This field is a member of `oneof`_ ``image_type``.
        container_image (google.events.cloud.notebooks_v1.types.ContainerImage):
            Use a container image to start the notebook
            instance.

            This field is a member of `oneof`_ ``image_type``.
        post_startup_script (str):
            Path to a Bash script that automatically runs after a
            notebook instance fully boots up. The path must be a URL or
            Cloud Storage path. Example:
            ``"gs://path-to-file/file-name"``
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The time at which this
            environment was created.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=2,
    )
    description: str = proto.Field(
        proto.STRING,
        number=3,
    )
    vm_image: 'VmImage' = proto.Field(
        proto.MESSAGE,
        number=6,
        oneof='image_type',
        message='VmImage',
    )
    container_image: 'ContainerImage' = proto.Field(
        proto.MESSAGE,
        number=7,
        oneof='image_type',
        message='ContainerImage',
    )
    post_startup_script: str = proto.Field(
        proto.STRING,
        number=8,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=9,
        message=timestamp_pb2.Timestamp,
    )


class VmImage(proto.Message):
    r"""Definition of a custom Compute Engine virtual machine image
    for starting a notebook instance with the environment installed
    directly on the VM.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        project (str):
            Required. The name of the Google Cloud project that this VM
            image belongs to. Format: ``{project_id}``
        image_name (str):
            Use VM image name to find the image.

            This field is a member of `oneof`_ ``image``.
        image_family (str):
            Use this VM image family to find the image;
            the newest image in this family will be used.

            This field is a member of `oneof`_ ``image``.
    """

    project: str = proto.Field(
        proto.STRING,
        number=1,
    )
    image_name: str = proto.Field(
        proto.STRING,
        number=2,
        oneof='image',
    )
    image_family: str = proto.Field(
        proto.STRING,
        number=3,
        oneof='image',
    )


class ContainerImage(proto.Message):
    r"""Definition of a container image for starting a notebook
    instance with the environment installed in a container.

    Attributes:
        repository (str):
            Required. The path to the container image repository. For
            example: ``gcr.io/{project_id}/{image_name}``
        tag (str):
            The tag of the container image. If not
            specified, this defaults to the latest tag.
    """

    repository: str = proto.Field(
        proto.STRING,
        number=1,
    )
    tag: str = proto.Field(
        proto.STRING,
        number=2,
    )


class Runtime(proto.Message):
    r"""The definition of a Runtime for a managed notebook instance.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        name (str):
            Output only. The resource name of the runtime. Format:
            ``projects/{project}/locations/{location}/runtimes/{runtimeId}``
        virtual_machine (google.events.cloud.notebooks_v1.types.VirtualMachine):
            Use a Compute Engine VM image to start the
            managed notebook instance.

            This field is a member of `oneof`_ ``runtime_type``.
        state (google.events.cloud.notebooks_v1.types.Runtime.State):
            Output only. Runtime state.
        health_state (google.events.cloud.notebooks_v1.types.Runtime.HealthState):
            Output only. Runtime health_state.
        access_config (google.events.cloud.notebooks_v1.types.RuntimeAccessConfig):
            The config settings for accessing runtime.
        software_config (google.events.cloud.notebooks_v1.types.RuntimeSoftwareConfig):
            The config settings for software inside the
            runtime.
        metrics (google.events.cloud.notebooks_v1.types.RuntimeMetrics):
            Output only. Contains Runtime daemon metrics
            such as Service status and JupyterLab stats.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Runtime creation time.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Runtime update time.
        labels (MutableMapping[str, str]):
            Optional. The labels to associate with this Managed Notebook
            or Runtime. Label **keys** must contain 1 to 63 characters,
            and must conform to `RFC
            1035 <https://www.ietf.org/rfc/rfc1035.txt>`__. Label
            **values** may be empty, but, if present, must contain 1 to
            63 characters, and must conform to `RFC
            1035 <https://www.ietf.org/rfc/rfc1035.txt>`__. No more than
            32 labels can be associated with a cluster.
    """
    class State(proto.Enum):
        r"""The definition of the states of this runtime.

        Values:
            STATE_UNSPECIFIED (0):
                State is not specified.
            STARTING (1):
                The compute layer is starting the runtime. It
                is not ready for use.
            PROVISIONING (2):
                The compute layer is installing required
                frameworks and registering the runtime with
                notebook proxy. It cannot be used.
            ACTIVE (3):
                The runtime is currently running. It is ready
                for use.
            STOPPING (4):
                The control logic is stopping the runtime. It
                cannot be used.
            STOPPED (5):
                The runtime is stopped. It cannot be used.
            DELETING (6):
                The runtime is being deleted. It cannot be
                used.
            UPGRADING (7):
                The runtime is upgrading. It cannot be used.
            INITIALIZING (8):
                The runtime is being created and set up. It
                is not ready for use.
        """
        STATE_UNSPECIFIED = 0
        STARTING = 1
        PROVISIONING = 2
        ACTIVE = 3
        STOPPING = 4
        STOPPED = 5
        DELETING = 6
        UPGRADING = 7
        INITIALIZING = 8

    class HealthState(proto.Enum):
        r"""The runtime substate.

        Values:
            HEALTH_STATE_UNSPECIFIED (0):
                The runtime substate is unknown.
            HEALTHY (1):
                The runtime is known to be in an healthy
                state (for example, critical daemons are
                running) Applies to ACTIVE state.
            UNHEALTHY (2):
                The runtime is known to be in an unhealthy
                state (for example, critical daemons are not
                running) Applies to ACTIVE state.
            AGENT_NOT_INSTALLED (3):
                The runtime has not installed health
                monitoring agent. Applies to ACTIVE state.
            AGENT_NOT_RUNNING (4):
                The runtime health monitoring agent is not
                running. Applies to ACTIVE state.
        """
        HEALTH_STATE_UNSPECIFIED = 0
        HEALTHY = 1
        UNHEALTHY = 2
        AGENT_NOT_INSTALLED = 3
        AGENT_NOT_RUNNING = 4

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    virtual_machine: 'VirtualMachine' = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof='runtime_type',
        message='VirtualMachine',
    )
    state: State = proto.Field(
        proto.ENUM,
        number=3,
        enum=State,
    )
    health_state: HealthState = proto.Field(
        proto.ENUM,
        number=4,
        enum=HealthState,
    )
    access_config: 'RuntimeAccessConfig' = proto.Field(
        proto.MESSAGE,
        number=5,
        message='RuntimeAccessConfig',
    )
    software_config: 'RuntimeSoftwareConfig' = proto.Field(
        proto.MESSAGE,
        number=6,
        message='RuntimeSoftwareConfig',
    )
    metrics: 'RuntimeMetrics' = proto.Field(
        proto.MESSAGE,
        number=7,
        message='RuntimeMetrics',
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=20,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=21,
        message=timestamp_pb2.Timestamp,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=23,
    )


class RuntimeAcceleratorConfig(proto.Message):
    r"""Definition of the types of hardware accelerators that can be used.
    Definition of the types of hardware accelerators that can be used.
    See `Compute Engine
    AcceleratorTypes <https://cloud.google.com/compute/docs/reference/beta/acceleratorTypes>`__.
    Examples:

    -  ``nvidia-tesla-k80``
    -  ``nvidia-tesla-p100``
    -  ``nvidia-tesla-v100``
    -  ``nvidia-tesla-p4``
    -  ``nvidia-tesla-t4``
    -  ``nvidia-tesla-a100``

    Attributes:
        type_ (google.events.cloud.notebooks_v1.types.RuntimeAcceleratorConfig.AcceleratorType):
            Accelerator model.
        core_count (int):
            Count of cores of this accelerator.
    """
    class AcceleratorType(proto.Enum):
        r"""Type of this accelerator.

        Values:
            ACCELERATOR_TYPE_UNSPECIFIED (0):
                Accelerator type is not specified.
            NVIDIA_TESLA_K80 (1):
                Accelerator type is Nvidia Tesla K80.
            NVIDIA_TESLA_P100 (2):
                Accelerator type is Nvidia Tesla P100.
            NVIDIA_TESLA_V100 (3):
                Accelerator type is Nvidia Tesla V100.
            NVIDIA_TESLA_P4 (4):
                Accelerator type is Nvidia Tesla P4.
            NVIDIA_TESLA_T4 (5):
                Accelerator type is Nvidia Tesla T4.
            NVIDIA_TESLA_A100 (6):
                Accelerator type is Nvidia Tesla A100 - 40GB.
            TPU_V2 (7):
                (Coming soon) Accelerator type is TPU V2.
            TPU_V3 (8):
                (Coming soon) Accelerator type is TPU V3.
            NVIDIA_TESLA_T4_VWS (9):
                Accelerator type is NVIDIA Tesla T4 Virtual
                Workstations.
            NVIDIA_TESLA_P100_VWS (10):
                Accelerator type is NVIDIA Tesla P100 Virtual
                Workstations.
            NVIDIA_TESLA_P4_VWS (11):
                Accelerator type is NVIDIA Tesla P4 Virtual
                Workstations.
        """
        ACCELERATOR_TYPE_UNSPECIFIED = 0
        NVIDIA_TESLA_K80 = 1
        NVIDIA_TESLA_P100 = 2
        NVIDIA_TESLA_V100 = 3
        NVIDIA_TESLA_P4 = 4
        NVIDIA_TESLA_T4 = 5
        NVIDIA_TESLA_A100 = 6
        TPU_V2 = 7
        TPU_V3 = 8
        NVIDIA_TESLA_T4_VWS = 9
        NVIDIA_TESLA_P100_VWS = 10
        NVIDIA_TESLA_P4_VWS = 11

    type_: AcceleratorType = proto.Field(
        proto.ENUM,
        number=1,
        enum=AcceleratorType,
    )
    core_count: int = proto.Field(
        proto.INT64,
        number=2,
    )


class EncryptionConfig(proto.Message):
    r"""Represents a custom encryption key configuration that can be
    applied to a resource. This will encrypt all disks in Virtual
    Machine.

    Attributes:
        kms_key (str):
            The Cloud KMS resource identifier of the customer-managed
            encryption key used to protect a resource, such as a disks.
            It has the following format:
            ``projects/{PROJECT_ID}/locations/{REGION}/keyRings/{KEY_RING_NAME}/cryptoKeys/{KEY_NAME}``
    """

    kms_key: str = proto.Field(
        proto.STRING,
        number=1,
    )


class LocalDisk(proto.Message):
    r"""A Local attached disk resource.

    Attributes:
        auto_delete (bool):
            Optional. Output only. Specifies whether the
            disk will be auto-deleted when the instance is
            deleted (but not when the disk is detached from
            the instance).
        boot (bool):
            Optional. Output only. Indicates that this is
            a boot disk. The virtual machine will use the
            first partition of the disk for its root
            filesystem.
        device_name (str):
            Optional. Output only. Specifies a unique device name of
            your choice that is reflected into the
            ``/dev/disk/by-id/google-*`` tree of a Linux operating
            system running within the instance. This name can be used to
            reference the device for mounting, resizing, and so on, from
            within the instance.

            If not specified, the server chooses a default device name
            to apply to this disk, in the form persistent-disk-x, where
            x is a number assigned by Google Compute Engine. This field
            is only applicable for persistent disks.
        guest_os_features (MutableSequence[google.events.cloud.notebooks_v1.types.LocalDisk.RuntimeGuestOsFeature]):
            Output only. Indicates a list of features to
            enable on the guest operating system. Applicable
            only for bootable images. Read  Enabling guest
            operating system features to see a list of
            available options.
        index (int):
            Output only. A zero-based index to this disk,
            where 0 is reserved for the boot disk. If you
            have many disks attached to an instance, each
            disk would have a unique index number.
        interface (str):
            Specifies the disk interface to use for attaching this disk,
            which is either SCSI or NVME. The default is SCSI.
            Persistent disks must always use SCSI and the request will
            fail if you attempt to attach a persistent disk in any other
            format than SCSI. Local SSDs can use either NVME or SCSI.
            For performance characteristics of SCSI over NVMe, see Local
            SSD performance. Valid values:

            -  ``NVME``
            -  ``SCSI``
        kind (str):
            Output only. Type of the resource. Always
            compute#attachedDisk for attached disks.
        licenses (MutableSequence[str]):
            Output only. Any valid publicly visible
            licenses.
        mode (str):
            The mode in which to attach this disk, either ``READ_WRITE``
            or ``READ_ONLY``. If not specified, the default is to attach
            the disk in ``READ_WRITE`` mode. Valid values:

            -  ``READ_ONLY``
            -  ``READ_WRITE``
        source (str):
            Specifies a valid partial or full URL to an
            existing Persistent Disk resource.
        type_ (str):
            Specifies the type of the disk, either ``SCRATCH`` or
            ``PERSISTENT``. If not specified, the default is
            ``PERSISTENT``. Valid values:

            -  ``PERSISTENT``
            -  ``SCRATCH``
    """

    class RuntimeGuestOsFeature(proto.Message):
        r"""Optional. A list of features to enable on the guest operating
        system. Applicable only for bootable images. Read `Enabling guest
        operating system
        features <https://cloud.google.com/compute/docs/images/create-delete-deprecate-private-images#guest-os-features>`__
        to see a list of available options. Guest OS features for boot disk.

        Attributes:
            type_ (str):
                The ID of a supported feature. Read `Enabling guest
                operating system
                features <https://cloud.google.com/compute/docs/images/create-delete-deprecate-private-images#guest-os-features>`__
                to see a list of available options.

                Valid values:

                -  ``FEATURE_TYPE_UNSPECIFIED``
                -  ``MULTI_IP_SUBNET``
                -  ``SECURE_BOOT``
                -  ``UEFI_COMPATIBLE``
                -  ``VIRTIO_SCSI_MULTIQUEUE``
                -  ``WINDOWS``
        """

        type_: str = proto.Field(
            proto.STRING,
            number=1,
        )

    auto_delete: bool = proto.Field(
        proto.BOOL,
        number=1,
    )
    boot: bool = proto.Field(
        proto.BOOL,
        number=2,
    )
    device_name: str = proto.Field(
        proto.STRING,
        number=3,
    )
    guest_os_features: MutableSequence[RuntimeGuestOsFeature] = proto.RepeatedField(
        proto.MESSAGE,
        number=4,
        message=RuntimeGuestOsFeature,
    )
    index: int = proto.Field(
        proto.INT32,
        number=5,
    )
    interface: str = proto.Field(
        proto.STRING,
        number=7,
    )
    kind: str = proto.Field(
        proto.STRING,
        number=8,
    )
    licenses: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=9,
    )
    mode: str = proto.Field(
        proto.STRING,
        number=10,
    )
    source: str = proto.Field(
        proto.STRING,
        number=11,
    )
    type_: str = proto.Field(
        proto.STRING,
        number=12,
    )


class LocalDiskInitializeParams(proto.Message):
    r"""Input only. Specifies the parameters for a new disk that will
    be created alongside the new instance. Use initialization
    parameters to create boot disks or local SSDs attached to the
    new runtime.
    This property is mutually exclusive with the source property;
    you can only define one or the other, but not both.

    Attributes:
        description (str):
            Optional. Provide this property when creating
            the disk.
        disk_name (str):
            Optional. Specifies the disk name. If not
            specified, the default is to use the name of the
            instance. If the disk with the instance name
            exists already in the given zone/region, a new
            name will be automatically generated.
        disk_size_gb (int):
            Optional. Specifies the size of the disk in
            base-2 GB. If not specified, the disk will be
            the same size as the image (usually 10GB). If
            specified, the size must be equal to or larger
            than 10GB. Default 100 GB.
        labels (MutableMapping[str, str]):
            Optional. Labels to apply to this disk. These
            can be later modified by the disks.setLabels
            method. This field is only applicable for
            persistent disks.
    """
    class DiskType(proto.Enum):
        r"""Possible disk types.

        Values:
            DISK_TYPE_UNSPECIFIED (0):
                Disk type not set.
            PD_STANDARD (1):
                Standard persistent disk type.
            PD_SSD (2):
                SSD persistent disk type.
            PD_BALANCED (3):
                Balanced persistent disk type.
            PD_EXTREME (4):
                Extreme persistent disk type.
        """
        DISK_TYPE_UNSPECIFIED = 0
        PD_STANDARD = 1
        PD_SSD = 2
        PD_BALANCED = 3
        PD_EXTREME = 4

    description: str = proto.Field(
        proto.STRING,
        number=1,
    )
    disk_name: str = proto.Field(
        proto.STRING,
        number=2,
    )
    disk_size_gb: int = proto.Field(
        proto.INT64,
        number=3,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=5,
    )


class RuntimeAccessConfig(proto.Message):
    r"""Specifies the login configuration for Runtime

    Attributes:
        access_type (google.events.cloud.notebooks_v1.types.RuntimeAccessConfig.RuntimeAccessType):
            The type of access mode this instance.
        runtime_owner (str):
            The owner of this runtime after creation. Format:
            ``alias@example.com`` Currently supports one owner only.
        proxy_uri (str):
            Output only. The proxy endpoint that is used
            to access the runtime.
    """
    class RuntimeAccessType(proto.Enum):
        r"""Possible ways to access runtime. Authentication mode.
        Currently supports: Single User only.

        Values:
            RUNTIME_ACCESS_TYPE_UNSPECIFIED (0):
                Unspecified access.
            SINGLE_USER (1):
                Single user login.
            SERVICE_ACCOUNT (2):
                Service Account mode.
                In Service Account mode, Runtime creator will
                specify a SA that exists in the consumer
                project. Using Runtime Service Account field.
                Users accessing the Runtime need ActAs (Service
                Account User) permission.
        """
        RUNTIME_ACCESS_TYPE_UNSPECIFIED = 0
        SINGLE_USER = 1
        SERVICE_ACCOUNT = 2

    access_type: RuntimeAccessType = proto.Field(
        proto.ENUM,
        number=1,
        enum=RuntimeAccessType,
    )
    runtime_owner: str = proto.Field(
        proto.STRING,
        number=2,
    )
    proxy_uri: str = proto.Field(
        proto.STRING,
        number=3,
    )


class RuntimeSoftwareConfig(proto.Message):
    r"""Specifies the selection and configuration of software inside the
    runtime. The properties to set on runtime. Properties keys are
    specified in ``key:value`` format, for example:

    -  ``idle_shutdown: true``
    -  ``idle_shutdown_timeout: 180``
    -  ``enable_health_monitoring: true``


    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        notebook_upgrade_schedule (str):
            Cron expression in UTC timezone, used to schedule instance
            auto upgrade. Please follow the `cron
            format <https://en.wikipedia.org/wiki/Cron>`__.
        enable_health_monitoring (bool):
            Verifies core internal services are running.
            Default: True

            This field is a member of `oneof`_ ``_enable_health_monitoring``.
        idle_shutdown (bool):
            Runtime will automatically shutdown after
            idle_shutdown_time. Default: True

            This field is a member of `oneof`_ ``_idle_shutdown``.
        idle_shutdown_timeout (int):
            Time in minutes to wait before shutting down
            runtime. Default: 180 minutes
        install_gpu_driver (bool):
            Install Nvidia Driver automatically.
            Default: True
        custom_gpu_driver_path (str):
            Specify a custom Cloud Storage path where the
            GPU driver is stored. If not specified, we'll
            automatically choose from official GPU drivers.
        post_startup_script (str):
            Path to a Bash script that automatically runs after a
            notebook instance fully boots up. The path must be a URL or
            Cloud Storage path (``gs://path-to-file/file-name``).
        kernels (MutableSequence[google.events.cloud.notebooks_v1.types.ContainerImage]):
            Optional. Use a list of container images to
            use as Kernels in the notebook instance.
        upgradeable (bool):
            Output only. Bool indicating whether an newer
            image is available in an image family.

            This field is a member of `oneof`_ ``_upgradeable``.
        post_startup_script_behavior (google.events.cloud.notebooks_v1.types.RuntimeSoftwareConfig.PostStartupScriptBehavior):
            Behavior for the post startup script.
        disable_terminal (bool):
            Bool indicating whether JupyterLab terminal
            will be available or not. Default: False

            This field is a member of `oneof`_ ``_disable_terminal``.
        version (str):
            Output only. version of boot image such as
            M100, from release label of the image.

            This field is a member of `oneof`_ ``_version``.
        mixer_disabled (bool):
            Bool indicating whether mixer client should
            be disabled. Default: False

            This field is a member of `oneof`_ ``_mixer_disabled``.
    """
    class PostStartupScriptBehavior(proto.Enum):
        r"""Behavior for the post startup script.

        Values:
            POST_STARTUP_SCRIPT_BEHAVIOR_UNSPECIFIED (0):
                Unspecified post startup script behavior.
                Will run only once at creation.
            RUN_EVERY_START (1):
                Runs the post startup script provided during
                creation at every start.
            DOWNLOAD_AND_RUN_EVERY_START (2):
                Downloads and runs the provided post startup
                script at every start.
        """
        POST_STARTUP_SCRIPT_BEHAVIOR_UNSPECIFIED = 0
        RUN_EVERY_START = 1
        DOWNLOAD_AND_RUN_EVERY_START = 2

    notebook_upgrade_schedule: str = proto.Field(
        proto.STRING,
        number=1,
    )
    enable_health_monitoring: bool = proto.Field(
        proto.BOOL,
        number=2,
        optional=True,
    )
    idle_shutdown: bool = proto.Field(
        proto.BOOL,
        number=3,
        optional=True,
    )
    idle_shutdown_timeout: int = proto.Field(
        proto.INT32,
        number=4,
    )
    install_gpu_driver: bool = proto.Field(
        proto.BOOL,
        number=5,
    )
    custom_gpu_driver_path: str = proto.Field(
        proto.STRING,
        number=6,
    )
    post_startup_script: str = proto.Field(
        proto.STRING,
        number=7,
    )
    kernels: MutableSequence['ContainerImage'] = proto.RepeatedField(
        proto.MESSAGE,
        number=8,
        message='ContainerImage',
    )
    upgradeable: bool = proto.Field(
        proto.BOOL,
        number=9,
        optional=True,
    )
    post_startup_script_behavior: PostStartupScriptBehavior = proto.Field(
        proto.ENUM,
        number=10,
        enum=PostStartupScriptBehavior,
    )
    disable_terminal: bool = proto.Field(
        proto.BOOL,
        number=11,
        optional=True,
    )
    version: str = proto.Field(
        proto.STRING,
        number=12,
        optional=True,
    )
    mixer_disabled: bool = proto.Field(
        proto.BOOL,
        number=13,
        optional=True,
    )


class RuntimeMetrics(proto.Message):
    r"""Contains runtime daemon metrics, such as OS and kernels and
    sessions stats.

    Attributes:
        system_metrics (MutableMapping[str, str]):
            Output only. The system metrics.
    """

    system_metrics: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=1,
    )


class RuntimeShieldedInstanceConfig(proto.Message):
    r"""A set of Shielded Instance options. See `Images using supported
    Shielded VM
    features <https://cloud.google.com/compute/docs/instances/modifying-shielded-vm>`__.
    Not all combinations are valid.

    Attributes:
        enable_secure_boot (bool):
            Defines whether the instance has Secure Boot
            enabled.
            Secure Boot helps ensure that the system only
            runs authentic software by verifying the digital
            signature of all boot components, and halting
            the boot process if signature verification
            fails. Disabled by default.
        enable_vtpm (bool):
            Defines whether the instance has the vTPM
            enabled. Enabled by default.
        enable_integrity_monitoring (bool):
            Defines whether the instance has integrity
            monitoring enabled.
            Enables monitoring and attestation of the boot
            integrity of the instance. The attestation is
            performed against the integrity policy baseline.
            This baseline is initially derived from the
            implicitly trusted boot image when the instance
            is created. Enabled by default.
    """

    enable_secure_boot: bool = proto.Field(
        proto.BOOL,
        number=1,
    )
    enable_vtpm: bool = proto.Field(
        proto.BOOL,
        number=2,
    )
    enable_integrity_monitoring: bool = proto.Field(
        proto.BOOL,
        number=3,
    )


class VirtualMachine(proto.Message):
    r"""Runtime using Virtual Machine for computing.

    Attributes:
        instance_name (str):
            Output only. The user-friendly name of the
            Managed Compute Engine instance.
        instance_id (str):
            Output only. The unique identifier of the
            Managed Compute Engine instance.
        virtual_machine_config (google.events.cloud.notebooks_v1.types.VirtualMachineConfig):
            Virtual Machine configuration settings.
    """

    instance_name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    instance_id: str = proto.Field(
        proto.STRING,
        number=2,
    )
    virtual_machine_config: 'VirtualMachineConfig' = proto.Field(
        proto.MESSAGE,
        number=3,
        message='VirtualMachineConfig',
    )


class VirtualMachineConfig(proto.Message):
    r"""The config settings for virtual machine.

    Attributes:
        zone (str):
            Output only. The zone where the virtual machine is located.
            If using regional request, the notebooks service will pick a
            location in the corresponding runtime region. On a get
            request, zone will always be present. Example:

            -  ``us-central1-b``
        machine_type (str):
            Required. The Compute Engine machine type used for runtimes.
            Short name is valid. Examples:

            -  ``n1-standard-2``
            -  ``e2-standard-8``
        container_images (MutableSequence[google.events.cloud.notebooks_v1.types.ContainerImage]):
            Optional. Use a list of container images to
            use as Kernels in the notebook instance.
        data_disk (google.events.cloud.notebooks_v1.types.LocalDisk):
            Required. Data disk option configuration
            settings.
        encryption_config (google.events.cloud.notebooks_v1.types.EncryptionConfig):
            Optional. Encryption settings for virtual
            machine data disk.
        shielded_instance_config (google.events.cloud.notebooks_v1.types.RuntimeShieldedInstanceConfig):
            Optional. Shielded VM Instance configuration
            settings.
        accelerator_config (google.events.cloud.notebooks_v1.types.RuntimeAcceleratorConfig):
            Optional. The Compute Engine accelerator
            configuration for this runtime.
        network (str):
            Optional. The Compute Engine network to be used for machine
            communications. Cannot be specified with subnetwork. If
            neither ``network`` nor ``subnet`` is specified, the
            "default" network of the project is used, if it exists.

            A full URL or partial URI. Examples:

            -  ``https://www.googleapis.com/compute/v1/projects/[project_id]/global/networks/default``
            -  ``projects/[project_id]/global/networks/default``

            Runtimes are managed resources inside Google Infrastructure.
            Runtimes support the following network configurations:

            -  Google Managed Network (Network & subnet are empty)
            -  Consumer Project VPC (network & subnet are required).
               Requires configuring Private Service Access.
            -  Shared VPC (network & subnet are required). Requires
               configuring Private Service Access.
        subnet (str):
            Optional. The Compute Engine subnetwork to be used for
            machine communications. Cannot be specified with network.

            A full URL or partial URI are valid. Examples:

            -  ``https://www.googleapis.com/compute/v1/projects/[project_id]/regions/us-east1/subnetworks/sub0``
            -  ``projects/[project_id]/regions/us-east1/subnetworks/sub0``
        internal_ip_only (bool):
            Optional. If true, runtime will only have internal IP
            addresses. By default, runtimes are not restricted to
            internal IP addresses, and will have ephemeral external IP
            addresses assigned to each vm. This ``internal_ip_only``
            restriction can only be enabled for subnetwork enabled
            networks, and all dependencies must be configured to be
            accessible without external IP addresses.
        tags (MutableSequence[str]):
            Optional. The Compute Engine tags to add to runtime (see
            `Tagging
            instances <https://cloud.google.com/compute/docs/label-or-tag-resources#tags>`__).
        guest_attributes (MutableMapping[str, str]):
            Output only. The Compute Engine guest attributes. (see
            `Project and instance guest
            attributes <https://cloud.google.com/compute/docs/storing-retrieving-metadata#guest_attributes>`__).
        metadata (MutableMapping[str, str]):
            Optional. The Compute Engine metadata entries to add to
            virtual machine. (see `Project and instance
            metadata <https://cloud.google.com/compute/docs/storing-retrieving-metadata#project_and_instance_metadata>`__).
        labels (MutableMapping[str, str]):
            Optional. The labels to associate with this runtime. Label
            **keys** must contain 1 to 63 characters, and must conform
            to `RFC 1035 <https://www.ietf.org/rfc/rfc1035.txt>`__.
            Label **values** may be empty, but, if present, must contain
            1 to 63 characters, and must conform to `RFC
            1035 <https://www.ietf.org/rfc/rfc1035.txt>`__. No more than
            32 labels can be associated with a cluster.
        nic_type (google.events.cloud.notebooks_v1.types.VirtualMachineConfig.NicType):
            Optional. The type of vNIC to be used on this
            interface. This may be gVNIC or VirtioNet.
        reserved_ip_range (str):
            Optional. Reserved IP Range name is used for VPC Peering.
            The subnetwork allocation will use the range *name* if it's
            assigned.

            Example: managed-notebooks-range-c

            ::

                PEERING_RANGE_NAME_3=managed-notebooks-range-c
                gcloud compute addresses create $PEERING_RANGE_NAME_3 \
                  --global \
                  --prefix-length=24 \
                  --description="Google Cloud Managed Notebooks Range 24 c" \
                  --network=$NETWORK \
                  --addresses=192.168.0.0 \
                  --purpose=VPC_PEERING

            Field value will be: ``managed-notebooks-range-c``
        boot_image (google.events.cloud.notebooks_v1.types.VirtualMachineConfig.BootImage):
            Optional. Boot image metadata used for
            runtime upgradeability.
    """
    class NicType(proto.Enum):
        r"""The type of vNIC driver. Default should be UNSPECIFIED_NIC_TYPE.

        Values:
            UNSPECIFIED_NIC_TYPE (0):
                No type specified.
            VIRTIO_NET (1):
                VIRTIO
            GVNIC (2):
                GVNIC
        """
        UNSPECIFIED_NIC_TYPE = 0
        VIRTIO_NET = 1
        GVNIC = 2

    class BootImage(proto.Message):
        r"""Definition of the boot image used by the Runtime.
        Used to facilitate runtime upgradeability.

        """

    zone: str = proto.Field(
        proto.STRING,
        number=1,
    )
    machine_type: str = proto.Field(
        proto.STRING,
        number=2,
    )
    container_images: MutableSequence['ContainerImage'] = proto.RepeatedField(
        proto.MESSAGE,
        number=3,
        message='ContainerImage',
    )
    data_disk: 'LocalDisk' = proto.Field(
        proto.MESSAGE,
        number=4,
        message='LocalDisk',
    )
    encryption_config: 'EncryptionConfig' = proto.Field(
        proto.MESSAGE,
        number=5,
        message='EncryptionConfig',
    )
    shielded_instance_config: 'RuntimeShieldedInstanceConfig' = proto.Field(
        proto.MESSAGE,
        number=6,
        message='RuntimeShieldedInstanceConfig',
    )
    accelerator_config: 'RuntimeAcceleratorConfig' = proto.Field(
        proto.MESSAGE,
        number=7,
        message='RuntimeAcceleratorConfig',
    )
    network: str = proto.Field(
        proto.STRING,
        number=8,
    )
    subnet: str = proto.Field(
        proto.STRING,
        number=9,
    )
    internal_ip_only: bool = proto.Field(
        proto.BOOL,
        number=10,
    )
    tags: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=13,
    )
    guest_attributes: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=14,
    )
    metadata: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=15,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=16,
    )
    nic_type: NicType = proto.Field(
        proto.ENUM,
        number=17,
        enum=NicType,
    )
    reserved_ip_range: str = proto.Field(
        proto.STRING,
        number=18,
    )
    boot_image: BootImage = proto.Field(
        proto.MESSAGE,
        number=19,
        message=BootImage,
    )


class ExecutionTemplate(proto.Message):
    r"""The description a notebook execution workload.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        scale_tier (google.events.cloud.notebooks_v1.types.ExecutionTemplate.ScaleTier):
            Required. Scale tier of the hardware used for
            notebook execution. DEPRECATED Will be
            discontinued. As right now only CUSTOM is
            supported.
        master_type (str):
            Specifies the type of virtual machine to use for your
            training job's master worker. You must specify this field
            when ``scaleTier`` is set to ``CUSTOM``.

            You can use certain Compute Engine machine types directly in
            this field. The following types are supported:

            -  ``n1-standard-4``
            -  ``n1-standard-8``
            -  ``n1-standard-16``
            -  ``n1-standard-32``
            -  ``n1-standard-64``
            -  ``n1-standard-96``
            -  ``n1-highmem-2``
            -  ``n1-highmem-4``
            -  ``n1-highmem-8``
            -  ``n1-highmem-16``
            -  ``n1-highmem-32``
            -  ``n1-highmem-64``
            -  ``n1-highmem-96``
            -  ``n1-highcpu-16``
            -  ``n1-highcpu-32``
            -  ``n1-highcpu-64``
            -  ``n1-highcpu-96``

            Alternatively, you can use the following legacy machine
            types:

            -  ``standard``
            -  ``large_model``
            -  ``complex_model_s``
            -  ``complex_model_m``
            -  ``complex_model_l``
            -  ``standard_gpu``
            -  ``complex_model_m_gpu``
            -  ``complex_model_l_gpu``
            -  ``standard_p100``
            -  ``complex_model_m_p100``
            -  ``standard_v100``
            -  ``large_model_v100``
            -  ``complex_model_m_v100``
            -  ``complex_model_l_v100``

            Finally, if you want to use a TPU for training, specify
            ``cloud_tpu`` in this field. Learn more about the `special
            configuration options for training with
            TPU <https://cloud.google.com/ai-platform/training/docs/using-tpus#configuring_a_custom_tpu_machine>`__.
        accelerator_config (google.events.cloud.notebooks_v1.types.ExecutionTemplate.SchedulerAcceleratorConfig):
            Configuration (count and accelerator type)
            for hardware running notebook execution.
        labels (MutableMapping[str, str]):
            Labels for execution.
            If execution is scheduled, a field included will
            be 'nbs-scheduled'. Otherwise, it is an
            immediate execution, and an included field will
            be 'nbs-immediate'. Use fields to efficiently
            index between various types of executions.
        input_notebook_file (str):
            Path to the notebook file to execute. Must be in a Google
            Cloud Storage bucket. Format:
            ``gs://{bucket_name}/{folder}/{notebook_file_name}`` Ex:
            ``gs://notebook_user/scheduled_notebooks/sentiment_notebook.ipynb``
        container_image_uri (str):
            Container Image URI to a DLVM
            Example:
            'gcr.io/deeplearning-platform-release/base-cu100'
            More examples can be found at:

            https://cloud.google.com/ai-platform/deep-learning-containers/docs/choosing-container
        output_notebook_folder (str):
            Path to the notebook folder to write to. Must be in a Google
            Cloud Storage bucket path. Format:
            ``gs://{bucket_name}/{folder}`` Ex:
            ``gs://notebook_user/scheduled_notebooks``
        params_yaml_file (str):
            Parameters to be overridden in the notebook during
            execution. Ref
            https://papermill.readthedocs.io/en/latest/usage-parameterize.html
            on how to specifying parameters in the input notebook and
            pass them here in an YAML file. Ex:
            ``gs://notebook_user/scheduled_notebooks/sentiment_notebook_params.yaml``
        parameters (str):
            Parameters used within the 'input_notebook_file' notebook.
        service_account (str):
            The email address of a service account to use when running
            the execution. You must have the
            ``iam.serviceAccounts.actAs`` permission for the specified
            service account.
        job_type (google.events.cloud.notebooks_v1.types.ExecutionTemplate.JobType):
            The type of Job to be used on this execution.
        dataproc_parameters (google.events.cloud.notebooks_v1.types.ExecutionTemplate.DataprocParameters):
            Parameters used in Dataproc JobType
            executions.

            This field is a member of `oneof`_ ``job_parameters``.
        vertex_ai_parameters (google.events.cloud.notebooks_v1.types.ExecutionTemplate.VertexAIParameters):
            Parameters used in Vertex AI JobType
            executions.

            This field is a member of `oneof`_ ``job_parameters``.
        kernel_spec (str):
            Name of the kernel spec to use. This must be
            specified if the kernel spec name on the
            execution target does not match the name in the
            input notebook file.
        tensorboard (str):
            The name of a Vertex AI [Tensorboard] resource to which this
            execution will upload Tensorboard logs. Format:
            ``projects/{project}/locations/{location}/tensorboards/{tensorboard}``
    """
    class ScaleTier(proto.Enum):
        r"""Required. Specifies the machine types, the number of replicas
        for workers and parameter servers.

        Values:
            SCALE_TIER_UNSPECIFIED (0):
                Unspecified Scale Tier.
            BASIC (1):
                A single worker instance. This tier is
                suitable for learning how to use Cloud ML, and
                for experimenting with new models using small
                datasets.
            STANDARD_1 (2):
                Many workers and a few parameter servers.
            PREMIUM_1 (3):
                A large number of workers with many parameter
                servers.
            BASIC_GPU (4):
                A single worker instance with a K80 GPU.
            BASIC_TPU (5):
                A single worker instance with a Cloud TPU.
            CUSTOM (6):
                The CUSTOM tier is not a set tier, but rather enables you to
                use your own cluster specification. When you use this tier,
                set values to configure your processing cluster according to
                these guidelines:

                -  You *must* set ``ExecutionTemplate.masterType`` to
                   specify the type of machine to use for your master node.
                   This is the only required setting.
        """
        SCALE_TIER_UNSPECIFIED = 0
        BASIC = 1
        STANDARD_1 = 2
        PREMIUM_1 = 3
        BASIC_GPU = 4
        BASIC_TPU = 5
        CUSTOM = 6

    class SchedulerAcceleratorType(proto.Enum):
        r"""Hardware accelerator types for AI Platform Training jobs.

        Values:
            SCHEDULER_ACCELERATOR_TYPE_UNSPECIFIED (0):
                Unspecified accelerator type. Default to no
                GPU.
            NVIDIA_TESLA_K80 (1):
                Nvidia Tesla K80 GPU.
            NVIDIA_TESLA_P100 (2):
                Nvidia Tesla P100 GPU.
            NVIDIA_TESLA_V100 (3):
                Nvidia Tesla V100 GPU.
            NVIDIA_TESLA_P4 (4):
                Nvidia Tesla P4 GPU.
            NVIDIA_TESLA_T4 (5):
                Nvidia Tesla T4 GPU.
            NVIDIA_TESLA_A100 (10):
                Nvidia Tesla A100 GPU.
            TPU_V2 (6):
                TPU v2.
            TPU_V3 (7):
                TPU v3.
        """
        SCHEDULER_ACCELERATOR_TYPE_UNSPECIFIED = 0
        NVIDIA_TESLA_K80 = 1
        NVIDIA_TESLA_P100 = 2
        NVIDIA_TESLA_V100 = 3
        NVIDIA_TESLA_P4 = 4
        NVIDIA_TESLA_T4 = 5
        NVIDIA_TESLA_A100 = 10
        TPU_V2 = 6
        TPU_V3 = 7

    class JobType(proto.Enum):
        r"""The backend used for this execution.

        Values:
            JOB_TYPE_UNSPECIFIED (0):
                No type specified.
            VERTEX_AI (1):
                Custom Job in ``aiplatform.googleapis.com``. Default value
                for an execution.
            DATAPROC (2):
                Run execution on a cluster with Dataproc as a
                job.
                https://cloud.google.com/dataproc/docs/reference/rest/v1/projects.regions.jobs
        """
        JOB_TYPE_UNSPECIFIED = 0
        VERTEX_AI = 1
        DATAPROC = 2

    class SchedulerAcceleratorConfig(proto.Message):
        r"""Definition of a hardware accelerator. Note that not all combinations
        of ``type`` and ``core_count`` are valid. See `GPUs on Compute
        Engine <https://cloud.google.com/compute/docs/gpus>`__ to find a
        valid combination. TPUs are not supported.

        Attributes:
            type_ (google.events.cloud.notebooks_v1.types.ExecutionTemplate.SchedulerAcceleratorType):
                Type of this accelerator.
            core_count (int):
                Count of cores of this accelerator.
        """

        type_: 'ExecutionTemplate.SchedulerAcceleratorType' = proto.Field(
            proto.ENUM,
            number=1,
            enum='ExecutionTemplate.SchedulerAcceleratorType',
        )
        core_count: int = proto.Field(
            proto.INT64,
            number=2,
        )

    class DataprocParameters(proto.Message):
        r"""Parameters used in Dataproc JobType executions.

        Attributes:
            cluster (str):
                URI for cluster used to run Dataproc execution. Format:
                ``projects/{PROJECT_ID}/regions/{REGION}/clusters/{CLUSTER_NAME}``
        """

        cluster: str = proto.Field(
            proto.STRING,
            number=1,
        )

    class VertexAIParameters(proto.Message):
        r"""Parameters used in Vertex AI JobType executions.

        Attributes:
            network (str):
                The full name of the Compute Engine
                `network <https://cloud.google.com/compute/docs/networks-and-firewalls#networks>`__
                to which the Job should be peered. For example,
                ``projects/12345/global/networks/myVPC``.
                `Format <https://cloud.google.com/compute/docs/reference/rest/v1/networks/insert>`__
                is of the form
                ``projects/{project}/global/networks/{network}``. Where
                ``{project}`` is a project number, as in ``12345``, and
                ``{network}`` is a network name.

                Private services access must already be configured for the
                network. If left unspecified, the job is not peered with any
                network.
            env (MutableMapping[str, str]):
                Environment variables. At most 100 environment variables can
                be specified and unique. Example:
                ``GCP_BUCKET=gs://my-bucket/samples/``
        """

        network: str = proto.Field(
            proto.STRING,
            number=1,
        )
        env: MutableMapping[str, str] = proto.MapField(
            proto.STRING,
            proto.STRING,
            number=2,
        )

    scale_tier: ScaleTier = proto.Field(
        proto.ENUM,
        number=1,
        enum=ScaleTier,
    )
    master_type: str = proto.Field(
        proto.STRING,
        number=2,
    )
    accelerator_config: SchedulerAcceleratorConfig = proto.Field(
        proto.MESSAGE,
        number=3,
        message=SchedulerAcceleratorConfig,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=4,
    )
    input_notebook_file: str = proto.Field(
        proto.STRING,
        number=5,
    )
    container_image_uri: str = proto.Field(
        proto.STRING,
        number=6,
    )
    output_notebook_folder: str = proto.Field(
        proto.STRING,
        number=7,
    )
    params_yaml_file: str = proto.Field(
        proto.STRING,
        number=8,
    )
    parameters: str = proto.Field(
        proto.STRING,
        number=9,
    )
    service_account: str = proto.Field(
        proto.STRING,
        number=10,
    )
    job_type: JobType = proto.Field(
        proto.ENUM,
        number=11,
        enum=JobType,
    )
    dataproc_parameters: DataprocParameters = proto.Field(
        proto.MESSAGE,
        number=12,
        oneof='job_parameters',
        message=DataprocParameters,
    )
    vertex_ai_parameters: VertexAIParameters = proto.Field(
        proto.MESSAGE,
        number=13,
        oneof='job_parameters',
        message=VertexAIParameters,
    )
    kernel_spec: str = proto.Field(
        proto.STRING,
        number=14,
    )
    tensorboard: str = proto.Field(
        proto.STRING,
        number=15,
    )


class Execution(proto.Message):
    r"""The definition of a single executed notebook.

    Attributes:
        execution_template (google.events.cloud.notebooks_v1.types.ExecutionTemplate):
            execute metadata including name, hardware
            spec, region, labels, etc.
        name (str):
            Output only. The resource name of the execute. Format:
            ``projects/{project_id}/locations/{location}/executions/{execution_id}``
        display_name (str):
            Output only. Name used for UI purposes. Name can only
            contain alphanumeric characters and underscores '_'.
        description (str):
            A brief description of this execution.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Time the Execution was
            instantiated.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Time the Execution was last
            updated.
        state (google.events.cloud.notebooks_v1.types.Execution.State):
            Output only. State of the underlying AI
            Platform job.
        output_notebook_file (str):
            Output notebook file generated by this
            execution
        job_uri (str):
            Output only. The URI of the external job used
            to execute the notebook.
    """
    class State(proto.Enum):
        r"""Enum description of the state of the underlying AIP job.

        Values:
            STATE_UNSPECIFIED (0):
                The job state is unspecified.
            QUEUED (1):
                The job has been just created and processing
                has not yet begun.
            PREPARING (2):
                The service is preparing to execution the
                job.
            RUNNING (3):
                The job is in progress.
            SUCCEEDED (4):
                The job completed successfully.
            FAILED (5):
                The job failed. ``error_message`` should contain the details
                of the failure.
            CANCELLING (6):
                The job is being cancelled. ``error_message`` should
                describe the reason for the cancellation.
            CANCELLED (7):
                The job has been cancelled. ``error_message`` should
                describe the reason for the cancellation.
            EXPIRED (9):
                The job has become expired (relevant to
                Vertex AI jobs)
                https://cloud.google.com/vertex-ai/docs/reference/rest/v1/JobState
            INITIALIZING (10):
                The Execution is being created.
        """
        STATE_UNSPECIFIED = 0
        QUEUED = 1
        PREPARING = 2
        RUNNING = 3
        SUCCEEDED = 4
        FAILED = 5
        CANCELLING = 6
        CANCELLED = 7
        EXPIRED = 9
        INITIALIZING = 10

    execution_template: 'ExecutionTemplate' = proto.Field(
        proto.MESSAGE,
        number=1,
        message='ExecutionTemplate',
    )
    name: str = proto.Field(
        proto.STRING,
        number=2,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=3,
    )
    description: str = proto.Field(
        proto.STRING,
        number=4,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=5,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=6,
        message=timestamp_pb2.Timestamp,
    )
    state: State = proto.Field(
        proto.ENUM,
        number=7,
        enum=State,
    )
    output_notebook_file: str = proto.Field(
        proto.STRING,
        number=8,
    )
    job_uri: str = proto.Field(
        proto.STRING,
        number=9,
    )


class ReservationAffinity(proto.Message):
    r"""Reservation Affinity for consuming Zonal reservation.

    Attributes:
        consume_reservation_type (google.events.cloud.notebooks_v1.types.ReservationAffinity.Type):
            Optional. Type of reservation to consume
        key (str):
            Optional. Corresponds to the label key of
            reservation resource.
        values (MutableSequence[str]):
            Optional. Corresponds to the label values of
            reservation resource.
    """
    class Type(proto.Enum):
        r"""Indicates whether to consume capacity from an reservation or
        not.

        Values:
            TYPE_UNSPECIFIED (0):
                Default type.
            NO_RESERVATION (1):
                Do not consume from any allocated capacity.
            ANY_RESERVATION (2):
                Consume any reservation available.
            SPECIFIC_RESERVATION (3):
                Must consume from a specific reservation.
                Must specify key value fields for specifying the
                reservations.
        """
        TYPE_UNSPECIFIED = 0
        NO_RESERVATION = 1
        ANY_RESERVATION = 2
        SPECIFIC_RESERVATION = 3

    consume_reservation_type: Type = proto.Field(
        proto.ENUM,
        number=1,
        enum=Type,
    )
    key: str = proto.Field(
        proto.STRING,
        number=2,
    )
    values: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=3,
    )


class Instance(proto.Message):
    r"""The definition of a notebook instance.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        name (str):
            Output only. The name of this notebook instance. Format:
            ``projects/{project_id}/locations/{location}/instances/{instance_id}``
        vm_image (google.events.cloud.notebooks_v1.types.VmImage):
            Use a Compute Engine VM image to start the
            notebook instance.

            This field is a member of `oneof`_ ``environment``.
        container_image (google.events.cloud.notebooks_v1.types.ContainerImage):
            Use a container image to start the notebook
            instance.

            This field is a member of `oneof`_ ``environment``.
        post_startup_script (str):
            Path to a Bash script that automatically runs after a
            notebook instance fully boots up. The path must be a URL or
            Cloud Storage path (``gs://path-to-file/file-name``).
        proxy_uri (str):
            Output only. The proxy endpoint that is used
            to access the Jupyter notebook.
        service_account (str):
            The service account on this instance, giving access to other
            Google Cloud services. You can use any service account
            within the same project, but you must have the service
            account user permission to use the instance.

            If not specified, the `Compute Engine default service
            account <https://cloud.google.com/compute/docs/access/service-accounts#default_service_account>`__
            is used.
        service_account_scopes (MutableSequence[str]):
            Optional. The URIs of service account scopes to be included
            in Compute Engine instances.

            If not specified, the following
            `scopes <https://cloud.google.com/compute/docs/access/service-accounts#accesscopesiam>`__
            are defined:

            -  https://www.googleapis.com/auth/cloud-platform
            -  https://www.googleapis.com/auth/userinfo.email If not
               using default scopes, you need at least:
               https://www.googleapis.com/auth/compute
        machine_type (str):
            Required. The `Compute Engine machine
            type <https://cloud.google.com/compute/docs/machine-types>`__
            of this instance.
        accelerator_config (google.events.cloud.notebooks_v1.types.Instance.AcceleratorConfig):
            The hardware accelerator used on this instance. If you use
            accelerators, make sure that your configuration has `enough
            vCPUs and memory to support the ``machine_type`` you have
            selected <https://cloud.google.com/compute/docs/gpus/#gpus-list>`__.
        state (google.events.cloud.notebooks_v1.types.Instance.State):
            Output only. The state of this instance.
        install_gpu_driver (bool):
            Whether the end user authorizes Google Cloud
            to install GPU driver on this instance.
            If this field is empty or set to false, the GPU
            driver won't be installed. Only applicable to
            instances with GPUs.
        custom_gpu_driver_path (str):
            Specify a custom Cloud Storage path where the
            GPU driver is stored. If not specified, we'll
            automatically choose from official GPU drivers.
        disks (MutableSequence[google.events.cloud.notebooks_v1.types.Instance.Disk]):
            Output only. Attached disks to notebook
            instance.
        shielded_instance_config (google.events.cloud.notebooks_v1.types.Instance.ShieldedInstanceConfig):
            Optional. Shielded VM configuration. `Images using supported
            Shielded VM
            features <https://cloud.google.com/compute/docs/instances/modifying-shielded-vm>`__.
        no_public_ip (bool):
            If true, no external IP will be assigned to
            this instance.
        no_proxy_access (bool):
            If true, the notebook instance will not
            register with the proxy.
        network (str):
            The name of the VPC that this instance is in. Format:
            ``projects/{project_id}/global/networks/{network_id}``
        subnet (str):
            The name of the subnet that this instance is in. Format:
            ``projects/{project_id}/regions/{region}/subnetworks/{subnetwork_id}``
        labels (MutableMapping[str, str]):
            Labels to apply to this instance.
            These can be later modified by the setLabels
            method.
        metadata (MutableMapping[str, str]):
            Custom metadata to apply to this instance.
        tags (MutableSequence[str]):
            Optional. The Compute Engine tags to add to runtime (see
            `Tagging
            instances <https://cloud.google.com/compute/docs/label-or-tag-resources#tags>`__).
        upgrade_history (MutableSequence[google.events.cloud.notebooks_v1.types.Instance.UpgradeHistoryEntry]):
            The upgrade history of this instance.
        nic_type (google.events.cloud.notebooks_v1.types.Instance.NicType):
            Optional. The type of vNIC to be used on this
            interface. This may be gVNIC or VirtioNet.
        reservation_affinity (google.events.cloud.notebooks_v1.types.ReservationAffinity):
            Optional. The optional reservation affinity. Setting this
            field will apply the specified `Zonal Compute
            Reservation <https://cloud.google.com/compute/docs/instances/reserving-zonal-resources>`__
            to this notebook instance.
        creator (str):
            Output only. Email address of entity that
            sent original CreateInstance request.
        can_ip_forward (bool):
            Optional. Flag to enable ip forwarding or
            not, default false/off.
            https://cloud.google.com/vpc/docs/using-routes#canipforward
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Instance creation time.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Instance update time.
    """
    class AcceleratorType(proto.Enum):
        r"""Definition of the types of hardware accelerators that can be
        used on this instance.

        Values:
            ACCELERATOR_TYPE_UNSPECIFIED (0):
                Accelerator type is not specified.
            NVIDIA_TESLA_K80 (1):
                Accelerator type is Nvidia Tesla K80.
            NVIDIA_TESLA_P100 (2):
                Accelerator type is Nvidia Tesla P100.
            NVIDIA_TESLA_V100 (3):
                Accelerator type is Nvidia Tesla V100.
            NVIDIA_TESLA_P4 (4):
                Accelerator type is Nvidia Tesla P4.
            NVIDIA_TESLA_T4 (5):
                Accelerator type is Nvidia Tesla T4.
            NVIDIA_TESLA_A100 (11):
                Accelerator type is Nvidia Tesla A100.
            NVIDIA_TESLA_T4_VWS (8):
                Accelerator type is NVIDIA Tesla T4 Virtual
                Workstations.
            NVIDIA_TESLA_P100_VWS (9):
                Accelerator type is NVIDIA Tesla P100 Virtual
                Workstations.
            NVIDIA_TESLA_P4_VWS (10):
                Accelerator type is NVIDIA Tesla P4 Virtual
                Workstations.
            TPU_V2 (6):
                (Coming soon) Accelerator type is TPU V2.
            TPU_V3 (7):
                (Coming soon) Accelerator type is TPU V3.
        """
        ACCELERATOR_TYPE_UNSPECIFIED = 0
        NVIDIA_TESLA_K80 = 1
        NVIDIA_TESLA_P100 = 2
        NVIDIA_TESLA_V100 = 3
        NVIDIA_TESLA_P4 = 4
        NVIDIA_TESLA_T4 = 5
        NVIDIA_TESLA_A100 = 11
        NVIDIA_TESLA_T4_VWS = 8
        NVIDIA_TESLA_P100_VWS = 9
        NVIDIA_TESLA_P4_VWS = 10
        TPU_V2 = 6
        TPU_V3 = 7

    class State(proto.Enum):
        r"""The definition of the states of this instance.

        Values:
            STATE_UNSPECIFIED (0):
                State is not specified.
            STARTING (1):
                The control logic is starting the instance.
            PROVISIONING (2):
                The control logic is installing required
                frameworks and registering the instance with
                notebook proxy
            ACTIVE (3):
                The instance is running.
            STOPPING (4):
                The control logic is stopping the instance.
            STOPPED (5):
                The instance is stopped.
            DELETED (6):
                The instance is deleted.
            UPGRADING (7):
                The instance is upgrading.
            INITIALIZING (8):
                The instance is being created.
            REGISTERING (9):
                The instance is getting registered.
            SUSPENDING (10):
                The instance is suspending.
            SUSPENDED (11):
                The instance is suspended.
        """
        STATE_UNSPECIFIED = 0
        STARTING = 1
        PROVISIONING = 2
        ACTIVE = 3
        STOPPING = 4
        STOPPED = 5
        DELETED = 6
        UPGRADING = 7
        INITIALIZING = 8
        REGISTERING = 9
        SUSPENDING = 10
        SUSPENDED = 11

    class DiskType(proto.Enum):
        r"""Possible disk types for notebook instances.

        Values:
            DISK_TYPE_UNSPECIFIED (0):
                Disk type not set.
            PD_STANDARD (1):
                Standard persistent disk type.
            PD_SSD (2):
                SSD persistent disk type.
            PD_BALANCED (3):
                Balanced persistent disk type.
            PD_EXTREME (4):
                Extreme persistent disk type.
        """
        DISK_TYPE_UNSPECIFIED = 0
        PD_STANDARD = 1
        PD_SSD = 2
        PD_BALANCED = 3
        PD_EXTREME = 4

    class DiskEncryption(proto.Enum):
        r"""Definition of the disk encryption options.

        Values:
            DISK_ENCRYPTION_UNSPECIFIED (0):
                Disk encryption is not specified.
            GMEK (1):
                Use Google managed encryption keys to encrypt
                the boot disk.
            CMEK (2):
                Use customer managed encryption keys to
                encrypt the boot disk.
        """
        DISK_ENCRYPTION_UNSPECIFIED = 0
        GMEK = 1
        CMEK = 2

    class NicType(proto.Enum):
        r"""The type of vNIC driver. Default should be UNSPECIFIED_NIC_TYPE.

        Values:
            UNSPECIFIED_NIC_TYPE (0):
                No type specified.
            VIRTIO_NET (1):
                VIRTIO
            GVNIC (2):
                GVNIC
        """
        UNSPECIFIED_NIC_TYPE = 0
        VIRTIO_NET = 1
        GVNIC = 2

    class AcceleratorConfig(proto.Message):
        r"""Definition of a hardware accelerator. Note that not all combinations
        of ``type`` and ``core_count`` are valid. See `GPUs on Compute
        Engine <https://cloud.google.com/compute/docs/gpus/#gpus-list>`__ to
        find a valid combination. TPUs are not supported.

        Attributes:
            type_ (google.events.cloud.notebooks_v1.types.Instance.AcceleratorType):
                Type of this accelerator.
            core_count (int):
                Count of cores of this accelerator.
        """

        type_: 'Instance.AcceleratorType' = proto.Field(
            proto.ENUM,
            number=1,
            enum='Instance.AcceleratorType',
        )
        core_count: int = proto.Field(
            proto.INT64,
            number=2,
        )

    class Disk(proto.Message):
        r"""An instance-attached disk resource.

        Attributes:
            auto_delete (bool):
                Indicates whether the disk will be
                auto-deleted when the instance is deleted (but
                not when the disk is detached from the
                instance).
            boot (bool):
                Indicates that this is a boot disk. The
                virtual machine will use the first partition of
                the disk for its root filesystem.
            device_name (str):
                Indicates a unique device name of your choice that is
                reflected into the ``/dev/disk/by-id/google-*`` tree of a
                Linux operating system running within the instance. This
                name can be used to reference the device for mounting,
                resizing, and so on, from within the instance.

                If not specified, the server chooses a default device name
                to apply to this disk, in the form persistent-disk-x, where
                x is a number assigned by Google Compute Engine.This field
                is only applicable for persistent disks.
            disk_size_gb (int):
                Indicates the size of the disk in base-2 GB.
            guest_os_features (MutableSequence[google.events.cloud.notebooks_v1.types.Instance.Disk.GuestOsFeature]):
                Indicates a list of features to enable on the
                guest operating system. Applicable only for
                bootable images. Read  Enabling guest operating
                system features to see a list of available
                options.
            index (int):
                A zero-based index to this disk, where 0 is
                reserved for the boot disk. If you have many
                disks attached to an instance, each disk would
                have a unique index number.
            interface (str):
                Indicates the disk interface to use for attaching this disk,
                which is either SCSI or NVME. The default is SCSI.
                Persistent disks must always use SCSI and the request will
                fail if you attempt to attach a persistent disk in any other
                format than SCSI. Local SSDs can use either NVME or SCSI.
                For performance characteristics of SCSI over NVMe, see Local
                SSD performance. Valid values:

                -  ``NVME``
                -  ``SCSI``
            kind (str):
                Type of the resource. Always
                compute#attachedDisk for attached disks.
            licenses (MutableSequence[str]):
                A list of publicly visible licenses. Reserved
                for Google's use. A License represents billing
                and aggregate usage data for public and
                marketplace images.
            mode (str):
                The mode in which to attach this disk, either ``READ_WRITE``
                or ``READ_ONLY``. If not specified, the default is to attach
                the disk in ``READ_WRITE`` mode. Valid values:

                -  ``READ_ONLY``
                -  ``READ_WRITE``
            source (str):
                Indicates a valid partial or full URL to an
                existing Persistent Disk resource.
            type_ (str):
                Indicates the type of the disk, either ``SCRATCH`` or
                ``PERSISTENT``. Valid values:

                -  ``PERSISTENT``
                -  ``SCRATCH``
        """

        class GuestOsFeature(proto.Message):
            r"""Guest OS features for boot disk.

            Attributes:
                type_ (str):
                    The ID of a supported feature. Read Enabling guest operating
                    system features to see a list of available options. Valid
                    values:

                    -  ``FEATURE_TYPE_UNSPECIFIED``
                    -  ``MULTI_IP_SUBNET``
                    -  ``SECURE_BOOT``
                    -  ``UEFI_COMPATIBLE``
                    -  ``VIRTIO_SCSI_MULTIQUEUE``
                    -  ``WINDOWS``
            """

            type_: str = proto.Field(
                proto.STRING,
                number=1,
            )

        auto_delete: bool = proto.Field(
            proto.BOOL,
            number=1,
        )
        boot: bool = proto.Field(
            proto.BOOL,
            number=2,
        )
        device_name: str = proto.Field(
            proto.STRING,
            number=3,
        )
        disk_size_gb: int = proto.Field(
            proto.INT64,
            number=4,
        )
        guest_os_features: MutableSequence['Instance.Disk.GuestOsFeature'] = proto.RepeatedField(
            proto.MESSAGE,
            number=5,
            message='Instance.Disk.GuestOsFeature',
        )
        index: int = proto.Field(
            proto.INT64,
            number=6,
        )
        interface: str = proto.Field(
            proto.STRING,
            number=7,
        )
        kind: str = proto.Field(
            proto.STRING,
            number=8,
        )
        licenses: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=9,
        )
        mode: str = proto.Field(
            proto.STRING,
            number=10,
        )
        source: str = proto.Field(
            proto.STRING,
            number=11,
        )
        type_: str = proto.Field(
            proto.STRING,
            number=12,
        )

    class ShieldedInstanceConfig(proto.Message):
        r"""A set of Shielded Instance options. See `Images using supported
        Shielded VM
        features <https://cloud.google.com/compute/docs/instances/modifying-shielded-vm>`__.
        Not all combinations are valid.

        Attributes:
            enable_secure_boot (bool):
                Defines whether the instance has Secure Boot
                enabled.
                Secure Boot helps ensure that the system only
                runs authentic software by verifying the digital
                signature of all boot components, and halting
                the boot process if signature verification
                fails. Disabled by default.
            enable_vtpm (bool):
                Defines whether the instance has the vTPM
                enabled. Enabled by default.
            enable_integrity_monitoring (bool):
                Defines whether the instance has integrity
                monitoring enabled.
                Enables monitoring and attestation of the boot
                integrity of the instance. The attestation is
                performed against the integrity policy baseline.
                This baseline is initially derived from the
                implicitly trusted boot image when the instance
                is created. Enabled by default.
        """

        enable_secure_boot: bool = proto.Field(
            proto.BOOL,
            number=1,
        )
        enable_vtpm: bool = proto.Field(
            proto.BOOL,
            number=2,
        )
        enable_integrity_monitoring: bool = proto.Field(
            proto.BOOL,
            number=3,
        )

    class UpgradeHistoryEntry(proto.Message):
        r"""The entry of VM image upgrade history.

        Attributes:
            snapshot (str):
                The snapshot of the boot disk of this
                notebook instance before upgrade.
            vm_image (str):
                The VM image before this instance upgrade.
            container_image (str):
                The container image before this instance
                upgrade.
            framework (str):
                The framework of this notebook instance.
            version (str):
                The version of the notebook instance before
                this upgrade.
            state (google.events.cloud.notebooks_v1.types.Instance.UpgradeHistoryEntry.State):
                The state of this instance upgrade history
                entry.
            create_time (google.protobuf.timestamp_pb2.Timestamp):
                The time that this instance upgrade history
                entry is created.
            target_image (str):
                Target VM Image. Format:
                ``ainotebooks-vm/project/image-name/name``.
            action (google.events.cloud.notebooks_v1.types.Instance.UpgradeHistoryEntry.Action):
                Action. Rolloback or Upgrade.
            target_version (str):
                Target VM Version, like m63.
        """
        class State(proto.Enum):
            r"""The definition of the states of this upgrade history entry.

            Values:
                STATE_UNSPECIFIED (0):
                    State is not specified.
                STARTED (1):
                    The instance upgrade is started.
                SUCCEEDED (2):
                    The instance upgrade is succeeded.
                FAILED (3):
                    The instance upgrade is failed.
            """
            STATE_UNSPECIFIED = 0
            STARTED = 1
            SUCCEEDED = 2
            FAILED = 3

        class Action(proto.Enum):
            r"""The definition of operations of this upgrade history entry.

            Values:
                ACTION_UNSPECIFIED (0):
                    Operation is not specified.
                UPGRADE (1):
                    Upgrade.
                ROLLBACK (2):
                    Rollback.
            """
            ACTION_UNSPECIFIED = 0
            UPGRADE = 1
            ROLLBACK = 2

        snapshot: str = proto.Field(
            proto.STRING,
            number=1,
        )
        vm_image: str = proto.Field(
            proto.STRING,
            number=2,
        )
        container_image: str = proto.Field(
            proto.STRING,
            number=3,
        )
        framework: str = proto.Field(
            proto.STRING,
            number=4,
        )
        version: str = proto.Field(
            proto.STRING,
            number=5,
        )
        state: 'Instance.UpgradeHistoryEntry.State' = proto.Field(
            proto.ENUM,
            number=6,
            enum='Instance.UpgradeHistoryEntry.State',
        )
        create_time: timestamp_pb2.Timestamp = proto.Field(
            proto.MESSAGE,
            number=7,
            message=timestamp_pb2.Timestamp,
        )
        target_image: str = proto.Field(
            proto.STRING,
            number=8,
        )
        action: 'Instance.UpgradeHistoryEntry.Action' = proto.Field(
            proto.ENUM,
            number=9,
            enum='Instance.UpgradeHistoryEntry.Action',
        )
        target_version: str = proto.Field(
            proto.STRING,
            number=10,
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    vm_image: 'VmImage' = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof='environment',
        message='VmImage',
    )
    container_image: 'ContainerImage' = proto.Field(
        proto.MESSAGE,
        number=3,
        oneof='environment',
        message='ContainerImage',
    )
    post_startup_script: str = proto.Field(
        proto.STRING,
        number=4,
    )
    proxy_uri: str = proto.Field(
        proto.STRING,
        number=5,
    )
    service_account: str = proto.Field(
        proto.STRING,
        number=7,
    )
    service_account_scopes: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=31,
    )
    machine_type: str = proto.Field(
        proto.STRING,
        number=8,
    )
    accelerator_config: AcceleratorConfig = proto.Field(
        proto.MESSAGE,
        number=9,
        message=AcceleratorConfig,
    )
    state: State = proto.Field(
        proto.ENUM,
        number=10,
        enum=State,
    )
    install_gpu_driver: bool = proto.Field(
        proto.BOOL,
        number=11,
    )
    custom_gpu_driver_path: str = proto.Field(
        proto.STRING,
        number=12,
    )
    disks: MutableSequence[Disk] = proto.RepeatedField(
        proto.MESSAGE,
        number=28,
        message=Disk,
    )
    shielded_instance_config: ShieldedInstanceConfig = proto.Field(
        proto.MESSAGE,
        number=30,
        message=ShieldedInstanceConfig,
    )
    no_public_ip: bool = proto.Field(
        proto.BOOL,
        number=17,
    )
    no_proxy_access: bool = proto.Field(
        proto.BOOL,
        number=18,
    )
    network: str = proto.Field(
        proto.STRING,
        number=19,
    )
    subnet: str = proto.Field(
        proto.STRING,
        number=20,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=21,
    )
    metadata: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=22,
    )
    tags: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=32,
    )
    upgrade_history: MutableSequence[UpgradeHistoryEntry] = proto.RepeatedField(
        proto.MESSAGE,
        number=29,
        message=UpgradeHistoryEntry,
    )
    nic_type: NicType = proto.Field(
        proto.ENUM,
        number=33,
        enum=NicType,
    )
    reservation_affinity: 'ReservationAffinity' = proto.Field(
        proto.MESSAGE,
        number=34,
        message='ReservationAffinity',
    )
    creator: str = proto.Field(
        proto.STRING,
        number=36,
    )
    can_ip_forward: bool = proto.Field(
        proto.BOOL,
        number=39,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=23,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=24,
        message=timestamp_pb2.Timestamp,
    )


class Schedule(proto.Message):
    r"""The definition of a schedule.

    Attributes:
        name (str):
            Output only. The name of this schedule. Format:
            ``projects/{project_id}/locations/{location}/schedules/{schedule_id}``
        display_name (str):
            Output only. Display name used for UI purposes. Name can
            only contain alphanumeric characters, hyphens ``-``, and
            underscores ``_``.
        description (str):
            A brief description of this environment.
        state (google.events.cloud.notebooks_v1.types.Schedule.State):

        cron_schedule (str):
            Cron-tab formatted schedule by which the job will execute.
            Format: minute, hour, day of month, month, day of week, e.g.
            ``0 0 * * WED`` = every Wednesday More examples:
            https://crontab.guru/examples.html
        time_zone (str):
            Timezone on which the cron_schedule. The value of this field
            must be a time zone name from the tz database. TZ Database:
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones

            Note that some time zones include a provision for daylight
            savings time. The rules for daylight saving time are
            determined by the chosen tz. For UTC use the string "utc".
            If a time zone is not specified, the default will be in UTC
            (also known as GMT).
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Time the schedule was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Time the schedule was last
            updated.
        execution_template (google.events.cloud.notebooks_v1.types.ExecutionTemplate):
            Notebook Execution Template corresponding to
            this schedule.
        recent_executions (MutableSequence[google.events.cloud.notebooks_v1.types.Execution]):
            Output only. The most recent execution names
            triggered from this schedule and their
            corresponding states.
    """
    class State(proto.Enum):
        r"""State of the job.

        Values:
            STATE_UNSPECIFIED (0):
                Unspecified state.
            ENABLED (1):
                The job is executing normally.
            PAUSED (2):
                The job is paused by the user. It will not execute. A user
                can intentionally pause the job using [PauseJobRequest][].
            DISABLED (3):
                The job is disabled by the system due to
                error. The user cannot directly set a job to be
                disabled.
            UPDATE_FAILED (4):
                The job state resulting from a failed
                [CloudScheduler.UpdateJob][] operation. To recover a job
                from this state, retry [CloudScheduler.UpdateJob][] until a
                successful response is received.
            INITIALIZING (5):
                The schedule resource is being created.
            DELETING (6):
                The schedule resource is being deleted.
        """
        STATE_UNSPECIFIED = 0
        ENABLED = 1
        PAUSED = 2
        DISABLED = 3
        UPDATE_FAILED = 4
        INITIALIZING = 5
        DELETING = 6

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=2,
    )
    description: str = proto.Field(
        proto.STRING,
        number=3,
    )
    state: State = proto.Field(
        proto.ENUM,
        number=4,
        enum=State,
    )
    cron_schedule: str = proto.Field(
        proto.STRING,
        number=5,
    )
    time_zone: str = proto.Field(
        proto.STRING,
        number=6,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=7,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=8,
        message=timestamp_pb2.Timestamp,
    )
    execution_template: 'ExecutionTemplate' = proto.Field(
        proto.MESSAGE,
        number=9,
        message='ExecutionTemplate',
    )
    recent_executions: MutableSequence['Execution'] = proto.RepeatedField(
        proto.MESSAGE,
        number=10,
        message='Execution',
    )


class RuntimeEventData(proto.Message):
    r"""The data within all Runtime events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.notebooks_v1.types.Runtime):
            Optional. The Runtime event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'Runtime' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='Runtime',
    )


class ExecutionEventData(proto.Message):
    r"""The data within all Execution events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.notebooks_v1.types.Execution):
            Optional. The Execution event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'Execution' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='Execution',
    )


class InstanceEventData(proto.Message):
    r"""The data within all Instance events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.notebooks_v1.types.Instance):
            Optional. The Instance event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'Instance' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='Instance',
    )


class ScheduleEventData(proto.Message):
    r"""The data within all Schedule events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.notebooks_v1.types.Schedule):
            Optional. The Schedule event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'Schedule' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='Schedule',
    )


class EnvironmentEventData(proto.Message):
    r"""The data within all Environment events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.notebooks_v1.types.Environment):
            Optional. The Environment event payload.
            Unset for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'Environment' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='Environment',
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0


from typing import Optional, Protocol, runtime_checkable

from .....doc_utils import export_module
from .....import_utils import optional_import_block

with optional_import_block():
    from google.oauth2.credentials import Credentials


__all__ = ["GoogleCredentialsProvider"]


@runtime_checkable
@export_module("autogen.tools.experimental.google.authentication")
class GoogleCredentialsProvider(Protocol):
    """A protocol for Google credentials provider."""

    def get_credentials(self) -> Optional["Credentials"]:  # type: ignore[no-any-unimported]
        """Get the Google credentials."""
        ...

    @property
    def host(self) -> str:
        """The host from which to get the credentials."""
        ...

    @property
    def port(self) -> int:
        """The port from which to get the credentials."""
        ...



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

CICD_PIPELINE_NAME: Final = "cicd.pipeline.name"
"""
The human readable name of the pipeline within a CI/CD system.
"""

CICD_PIPELINE_RUN_ID: Final = "cicd.pipeline.run.id"
"""
The unique identifier of a pipeline run within a CI/CD system.
"""

CICD_PIPELINE_TASK_NAME: Final = "cicd.pipeline.task.name"
"""
The human readable name of a task within a pipeline. Task here most closely aligns with a [computing process](https://en.wikipedia.org/wiki/Pipeline_(computing)) in a pipeline. Other terms for tasks include commands, steps, and procedures.
"""

CICD_PIPELINE_TASK_RUN_ID: Final = "cicd.pipeline.task.run.id"
"""
The unique identifier of a task run within a pipeline.
"""

CICD_PIPELINE_TASK_RUN_URL_FULL: Final = "cicd.pipeline.task.run.url.full"
"""
The [URL](https://en.wikipedia.org/wiki/URL) of the pipeline run providing the complete address in order to locate and identify the pipeline run.
"""

CICD_PIPELINE_TASK_TYPE: Final = "cicd.pipeline.task.type"
"""
The type of the task within a pipeline.
"""


class CicdPipelineTaskTypeValues(Enum):
    BUILD = "build"
    """build."""
    TEST = "test"
    """test."""
    DEPLOY = "deploy"
    """deploy."""



from __future__ import annotations

import copy
import os
import pathlib
from io import BytesIO, StringIO
from typing import Any, Dict, Iterable, List, Optional, Tuple, TypedDict, cast

import requests
from langchain_core.documents import Document

from langchain_text_splitters.character import RecursiveCharacterTextSplitter


class ElementType(TypedDict):
    """Element type as typed dict."""

    url: str
    xpath: str
    content: str
    metadata: Dict[str, str]


class HTMLHeaderTextSplitter:
    """
    Splitting HTML files based on specified headers.
    Requires lxml package.
    """

    def __init__(
        self,
        headers_to_split_on: List[Tuple[str, str]],
        return_each_element: bool = False,
    ):
        """Create a new HTMLHeaderTextSplitter.

        Args:
            headers_to_split_on: list of tuples of headers we want to track mapped to
                (arbitrary) keys for metadata. Allowed header values: h1, h2, h3, h4,
                h5, h6 e.g. [("h1", "Header 1"), ("h2", "Header 2)].
            return_each_element: Return each element w/ associated headers.
        """
        # Output element-by-element or aggregated into chunks w/ common headers
        self.return_each_element = return_each_element
        self.headers_to_split_on = sorted(headers_to_split_on)

    def aggregate_elements_to_chunks(
        self, elements: List[ElementType]
    ) -> List[Document]:
        """Combine elements with common metadata into chunks

        Args:
            elements: HTML element content with associated identifying info and metadata
        """
        aggregated_chunks: List[ElementType] = []

        for element in elements:
            if (
                aggregated_chunks
                and aggregated_chunks[-1]["metadata"] == element["metadata"]
            ):
                # If the last element in the aggregated list
                # has the same metadata as the current element,
                # append the current content to the last element's content
                aggregated_chunks[-1]["content"] += "  \n" + element["content"]
            else:
                # Otherwise, append the current element to the aggregated list
                aggregated_chunks.append(element)

        return [
            Document(page_content=chunk["content"], metadata=chunk["metadata"])
            for chunk in aggregated_chunks
        ]

    def split_text_from_url(self, url: str) -> List[Document]:
        """Split HTML from web URL

        Args:
            url: web URL
        """
        r = requests.get(url)
        return self.split_text_from_file(BytesIO(r.content))

    def split_text(self, text: str) -> List[Document]:
        """Split HTML text string

        Args:
            text: HTML text
        """
        return self.split_text_from_file(StringIO(text))

    def split_text_from_file(self, file: Any) -> List[Document]:
        """Split HTML file

        Args:
            file: HTML file
        """
        try:
            from lxml import etree
        except ImportError as e:
            raise ImportError(
                "Unable to import lxml, please install with `pip install lxml`."
            ) from e
        # use lxml library to parse html document and return xml ElementTree
        # Explicitly encoding in utf-8 allows non-English
        # html files to be processed without garbled characters
        parser = etree.HTMLParser(encoding="utf-8")
        tree = etree.parse(file, parser)

        # document transformation for "structure-aware" chunking is handled with xsl.
        # see comments in html_chunks_with_headers.xslt for more detailed information.
        xslt_path = pathlib.Path(__file__).parent / "xsl/html_chunks_with_headers.xslt"
        xslt_tree = etree.parse(xslt_path)
        transform = etree.XSLT(xslt_tree)
        result = transform(tree)
        result_dom = etree.fromstring(str(result))

        # create filter and mapping for header metadata
        header_filter = [header[0] for header in self.headers_to_split_on]
        header_mapping = dict(self.headers_to_split_on)

        # map xhtml namespace prefix
        ns_map = {"h": "http://www.w3.org/1999/xhtml"}

        # build list of elements from DOM
        elements = []
        for element in result_dom.findall("*//*", ns_map):
            if element.findall("*[@class='headers']") or element.findall(
                "*[@class='chunk']"
            ):
                elements.append(
                    ElementType(
                        url=file,
                        xpath="".join(
                            [
                                node.text or ""
                                for node in element.findall("*[@class='xpath']", ns_map)
                            ]
                        ),
                        content="".join(
                            [
                                node.text or ""
                                for node in element.findall("*[@class='chunk']", ns_map)
                            ]
                        ),
                        metadata={
                            # Add text of specified headers to metadata using header
                            # mapping.
                            header_mapping[node.tag]: node.text or ""
                            for node in filter(
                                lambda x: x.tag in header_filter,
                                element.findall("*[@class='headers']/*", ns_map),
                            )
                        },
                    )
                )

        if not self.return_each_element:
            return self.aggregate_elements_to_chunks(elements)
        else:
            return [
                Document(page_content=chunk["content"], metadata=chunk["metadata"])
                for chunk in elements
            ]


class HTMLSectionSplitter:
    """
    Splitting HTML files based on specified tag and font sizes.
    Requires lxml package.
    """

    def __init__(
        self,
        headers_to_split_on: List[Tuple[str, str]],
        xslt_path: str = "xsl/converting_to_header.xslt",
        **kwargs: Any,
    ) -> None:
        """Create a new HTMLSectionSplitter.

        Args:
            headers_to_split_on: list of tuples of headers we want to track mapped to
                (arbitrary) keys for metadata. Allowed header values: h1, h2, h3, h4,
                h5, h6 e.g. [("h1", "Header 1"), ("h2", "Header 2"].
            xslt_path: path to xslt file for document transformation.
            Needed for html contents that using different format and layouts.
        """
        self.headers_to_split_on = dict(headers_to_split_on)
        self.xslt_path = xslt_path
        self.kwargs = kwargs

    def split_documents(self, documents: Iterable[Document]) -> List[Document]:
        """Split documents."""
        texts, metadatas = [], []
        for doc in documents:
            texts.append(doc.page_content)
            metadatas.append(doc.metadata)
        results = self.create_documents(texts, metadatas=metadatas)

        text_splitter = RecursiveCharacterTextSplitter(**self.kwargs)

        return text_splitter.split_documents(results)

    def split_text(self, text: str) -> List[Document]:
        """Split HTML text string

        Args:
            text: HTML text
        """
        return self.split_text_from_file(StringIO(text))

    def create_documents(
        self, texts: List[str], metadatas: Optional[List[dict]] = None
    ) -> List[Document]:
        """Create documents from a list of texts."""
        _metadatas = metadatas or [{}] * len(texts)
        documents = []
        for i, text in enumerate(texts):
            for chunk in self.split_text(text):
                metadata = copy.deepcopy(_metadatas[i])

                for key in chunk.metadata.keys():
                    if chunk.metadata[key] == "#TITLE#":
                        chunk.metadata[key] = metadata["Title"]
                metadata = {**metadata, **chunk.metadata}
                new_doc = Document(page_content=chunk.page_content, metadata=metadata)
                documents.append(new_doc)
        return documents

    def split_html_by_headers(
        self, html_doc: str
    ) -> Dict[str, Dict[str, Optional[str]]]:
        try:
            from bs4 import BeautifulSoup, PageElement  # type: ignore[import-untyped]
        except ImportError as e:
            raise ImportError(
                "Unable to import BeautifulSoup/PageElement, \
                    please install with `pip install \
                    bs4`."
            ) from e

        soup = BeautifulSoup(html_doc, "html.parser")
        headers = list(self.headers_to_split_on.keys())
        sections: Dict[str, Dict[str, Optional[str]]] = {}

        headers = soup.find_all(["body"] + headers)

        for i, header in enumerate(headers):
            header_element: PageElement = header
            if i == 0:
                current_header = "#TITLE#"
                current_header_tag = "h1"
                section_content: List = []
            else:
                current_header = header_element.text.strip()
                current_header_tag = header_element.name
                section_content = []
            for element in header_element.next_elements:
                if i + 1 < len(headers) and element == headers[i + 1]:
                    break
                if isinstance(element, str):
                    section_content.append(element)
            content = " ".join(section_content).strip()

            if content != "":
                sections[current_header] = {
                    "content": content,
                    "tag_name": current_header_tag,
                }

        return sections

    def convert_possible_tags_to_header(self, html_content: str) -> str:
        if self.xslt_path is None:
            return html_content

        try:
            from lxml import etree
        except ImportError as e:
            raise ImportError(
                "Unable to import lxml, please install with `pip install lxml`."
            ) from e
        # use lxml library to parse html document and return xml ElementTree
        parser = etree.HTMLParser()
        tree = etree.parse(StringIO(html_content), parser)

        # document transformation for "structure-aware" chunking is handled with xsl.
        # this is needed for htmls files that using different font sizes and layouts
        # check to see if self.xslt_path is a relative path or absolute path
        if not os.path.isabs(self.xslt_path):
            xslt_path = pathlib.Path(__file__).parent / self.xslt_path

        xslt_tree = etree.parse(xslt_path)
        transform = etree.XSLT(xslt_tree)
        result = transform(tree)
        return str(result)

    def split_text_from_file(self, file: Any) -> List[Document]:
        """Split HTML file

        Args:
            file: HTML file
        """
        file_content = file.getvalue()
        file_content = self.convert_possible_tags_to_header(file_content)
        sections = self.split_html_by_headers(file_content)

        return [
            Document(
                cast(str, sections[section_key]["content"]),
                metadata={
                    self.headers_to_split_on[
                        str(sections[section_key]["tag_name"])
                    ]: section_key
                },
            )
            for section_key in sections.keys()
        ]



# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Cloud functions to handle events from Firebase Alerts.
Subpackages give stronger typing to specific services which
notify users via Firebase Alerts.
"""

import dataclasses as _dataclasses
import datetime as _dt
import typing as _typing

from firebase_functions.core import T


@_dataclasses.dataclass(frozen=True)
class FirebaseAlertData(_typing.Generic[T]):
    """
    The CloudEvent data emitted by Firebase Alerts.
    """

    create_time: _dt.datetime
    """
    The time the alert was created.
    """

    end_time: _dt.datetime | None
    """
    The time the alert ended. This is only set for alerts that have ended.
    """

    payload: T
    """
    Payload of the event, which includes the details of the specific alert.
    """



import operator
from typing import Annotated, Sequence, TypedDict, Union

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.messages import BaseMessage
from langchain_core.runnables import RunnableLambda

from langgraph.graph import END, StateGraph
from langgraph.prebuilt.tool_executor import ToolExecutor


def _get_agent_state(input_schema=None):
    if input_schema is None:

        class AgentState(TypedDict):
            # The input string
            input: str
            # The list of previous messages in the conversation
            chat_history: Sequence[BaseMessage]
            # The outcome of a given call to the agent
            # Needs `None` as a valid type, since this is what this will start as
            agent_outcome: Union[AgentAction, AgentFinish, None]
            # List of actions and corresponding observations
            # Here we annotate this with `operator.add` to indicate that operations to
            # this state should be ADDED to the existing values (not overwrite it)
            intermediate_steps: Annotated[list[tuple[AgentAction, str]], operator.add]

    else:

        class AgentState(input_schema):
            # The outcome of a given call to the agent
            # Needs `None` as a valid type, since this is what this will start as
            agent_outcome: Union[AgentAction, AgentFinish, None]
            # List of actions and corresponding observations
            # Here we annotate this with `operator.add` to indicate that operations to
            # this state should be ADDED to the existing values (not overwrite it)
            intermediate_steps: Annotated[list[tuple[AgentAction, str]], operator.add]

    return AgentState


def create_agent_executor(agent_runnable, tools, input_schema=None):
    if isinstance(tools, ToolExecutor):
        tool_executor = tools
    else:
        tool_executor = ToolExecutor(tools)

    state = _get_agent_state(input_schema)

    # Define logic that will be used to determine which conditional edge to go down

    def should_continue(data):
        # If the agent outcome is an AgentFinish, then we return `exit` string
        # This will be used when setting up the graph to define the flow
        if isinstance(data["agent_outcome"], AgentFinish):
            return "end"
        # Otherwise, an AgentAction is returned
        # Here we return `continue` string
        # This will be used when setting up the graph to define the flow
        else:
            return "continue"

    def run_agent(data):
        agent_outcome = agent_runnable.invoke(data)
        return {"agent_outcome": agent_outcome}

    async def arun_agent(data):
        agent_outcome = await agent_runnable.ainvoke(data)
        return {"agent_outcome": agent_outcome}

    # Define the function to execute tools
    def execute_tools(data):
        # Get the most recent agent_outcome - this is the key added in the `agent` above
        agent_action = data["agent_outcome"]
        output = tool_executor.invoke(agent_action)
        return {"intermediate_steps": [(agent_action, str(output))]}

    async def aexecute_tools(data):
        # Get the most recent agent_outcome - this is the key added in the `agent` above
        agent_action = data["agent_outcome"]
        output = await tool_executor.ainvoke(agent_action)
        return {"intermediate_steps": [(agent_action, str(output))]}

    # Define a new graph
    workflow = StateGraph(state)

    # Define the two nodes we will cycle between
    workflow.add_node("agent", RunnableLambda(run_agent, arun_agent))
    workflow.add_node("action", RunnableLambda(execute_tools, aexecute_tools))

    # Set the entrypoint as `agent`
    # This means that this node is the first one called
    workflow.set_entry_point("agent")

    # We now add a conditional edge
    workflow.add_conditional_edges(
        # First, we define the start node. We use `agent`.
        # This means these are the edges taken after the `agent` node is called.
        "agent",
        # Next, we pass in the function that will determine which node is called next.
        should_continue,
        # Finally we pass in a mapping.
        # The keys are strings, and the values are other nodes.
        # END is a special node marking that the graph should finish.
        # What will happen is we will call `should_continue`, and then the output of that
        # will be matched against the keys in this mapping.
        # Based on which one it matches, that node will then be called.
        {
            # If `tools`, then we call the tool node.
            "continue": "action",
            # Otherwise we finish.
            "end": END,
        },
    )

    # We now add a normal edge from `tools` to `agent`.
    # This means that after `tools` is called, `agent` node is called next.
    workflow.add_edge("action", "agent")

    # Finally, we compile it!
    # This compiles it into a LangChain Runnable,
    # meaning you can use it as you would any other runnable
    return workflow.compile()



"""**Index** is used to avoid writing duplicated content
into the vectostore and to avoid over-writing content if it's unchanged.

Indexes also :

* Create knowledge graphs from data.

* Support indexing workflows from LangChain data loaders to vectorstores.

Importantly, Index keeps on working even if the content being written is derived
via a set of transformations from some source content (e.g., indexing children
documents that were derived from parent documents by chunking.)
"""



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

VCS_REPOSITORY_CHANGE_ID: Final = "vcs.repository.change.id"
"""
The ID of the change (pull request/merge request) if applicable. This is usually a unique (within repository) identifier generated by the VCS system.
"""

VCS_REPOSITORY_CHANGE_TITLE: Final = "vcs.repository.change.title"
"""
The human readable title of the change (pull request/merge request). This title is often a brief summary of the change and may get merged in to a ref as the commit summary.
"""

VCS_REPOSITORY_REF_NAME: Final = "vcs.repository.ref.name"
"""
The name of the [reference](https://git-scm.com/docs/gitglossary#def_ref) such as **branch** or **tag** in the repository.
"""

VCS_REPOSITORY_REF_REVISION: Final = "vcs.repository.ref.revision"
"""
The revision, literally [revised version](https://www.merriam-webster.com/dictionary/revision), The revision most often refers to a commit object in Git, or a revision number in SVN.
Note: The revision can be a full [hash value (see glossary)](https://nvlpubs.nist.gov/nistpubs/FIPS/NIST.FIPS.186-5.pdf),
of the recorded change to a ref within a repository pointing to a
commit [commit](https://git-scm.com/docs/git-commit) object. It does
not necessarily have to be a hash; it can simply define a
[revision number](https://svnbook.red-bean.com/en/1.7/svn.tour.revs.specifiers.html)
which is an integer that is monotonically increasing. In cases where
it is identical to the `ref.name`, it SHOULD still be included. It is
up to the implementer to decide which value to set as the revision
based on the VCS system and situational context.
"""

VCS_REPOSITORY_REF_TYPE: Final = "vcs.repository.ref.type"
"""
The type of the [reference](https://git-scm.com/docs/gitglossary#def_ref) in the repository.
"""

VCS_REPOSITORY_URL_FULL: Final = "vcs.repository.url.full"
"""
The [URL](https://en.wikipedia.org/wiki/URL) of the repository providing the complete address in order to locate and identify the repository.
"""


class VcsRepositoryRefTypeValues(Enum):
    BRANCH = "branch"
    """[branch](https://git-scm.com/docs/gitglossary#Documentation/gitglossary.txt-aiddefbranchabranch)."""
    TAG = "tag"
    """[tag](https://git-scm.com/docs/gitglossary#Documentation/gitglossary.txt-aiddeftagatag)."""



"""Tests for parse_from_object function and attribute docstrings."""
from unittest.mock import patch

from docstring_parser import parse_from_object

module_attr: int = 1
"""Description for module_attr"""


def test_from_module_attribute_docstrings() -> None:
    """Test the parse of attribute docstrings from a module."""
    from . import test_parse_from_object  # pylint: disable=C0415,W0406

    docstring = parse_from_object(test_parse_from_object)

    assert "parse_from_object" in docstring.short_description
    assert len(docstring.params) == 1
    assert docstring.params[0].arg_name == "module_attr"
    assert docstring.params[0].type_name == "int"
    assert docstring.params[0].description == "Description for module_attr"


def test_from_class_attribute_docstrings() -> None:
    """Test the parse of attribute docstrings from a class."""

    class StandardCase:
        """Short description
        Long description
        """

        attr_one: str
        """Description for attr_one"""
        attr_two: bool = False
        """Description for attr_two"""

    docstring = parse_from_object(StandardCase)

    assert docstring.short_description == "Short description"
    assert docstring.long_description == "Long description"
    assert len(docstring.params) == 2
    assert docstring.params[0].arg_name == "attr_one"
    assert docstring.params[0].type_name == "str"
    assert docstring.params[0].description == "Description for attr_one"
    assert docstring.params[1].arg_name == "attr_two"
    assert docstring.params[1].type_name == "bool"
    assert docstring.params[1].description == "Description for attr_two"


def test_from_class_attribute_docstrings_without_type() -> None:
    """Test the parse of untyped attribute docstrings."""

    class WithoutType:  # pylint: disable=missing-class-docstring
        attr_one = "value"
        """Description for attr_one"""

    docstring = parse_from_object(WithoutType)

    assert docstring.short_description is None
    assert docstring.long_description is None
    assert len(docstring.params) == 1
    assert docstring.params[0].arg_name == "attr_one"
    assert docstring.params[0].type_name is None
    assert docstring.params[0].description == "Description for attr_one"


def test_from_class_without_source() -> None:
    """Test the parse of class when source is unavailable."""

    class WithoutSource:
        """Short description"""

        attr_one: str
        """Description for attr_one"""

    with patch(
        "inspect.getsource", side_effect=OSError("could not get source code")
    ):
        docstring = parse_from_object(WithoutSource)

    assert docstring.short_description == "Short description"
    assert len(docstring.params) == 0


def test_from_function() -> None:
    """Test the parse of a function docstring."""

    def a_function(param1: str, param2: int = 2):
        """Short description
        Args:
            param1: Description for param1
            param2: Description for param2
        """
        return f"{param1} {param2}"

    docstring = parse_from_object(a_function)

    assert docstring.short_description == "Short description"
    assert len(docstring.params) == 2
    assert docstring.params[0].arg_name == "param1"
    assert docstring.params[0].type_name is None
    assert docstring.params[0].description == "Description for param1"
    assert docstring.params[1].arg_name == "param2"
    assert docstring.params[1].type_name is None
    assert docstring.params[1].description == "Description for param2"



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Final

FEATURE_FLAG_KEY: Final = "feature_flag.key"
"""
The unique identifier of the feature flag.
"""

FEATURE_FLAG_PROVIDER_NAME: Final = "feature_flag.provider_name"
"""
The name of the service provider that performs the flag evaluation.
"""

FEATURE_FLAG_VARIANT: Final = "feature_flag.variant"
"""
SHOULD be a semantic identifier for a value. If one is unavailable, a stringified version of the value can be used.
Note: A semantic identifier, commonly referred to as a variant, provides a means
for referring to a value without including the value itself. This can
provide additional context for understanding the meaning behind a value.
For example, the variant `red` maybe be used for the value `#c05543`.

A stringified version of the value can be used in situations where a
semantic identifier is unavailable. String representation of the value
should be determined by the implementer.
"""



"""Decorators for registering schema pre-processing and post-processing methods.
These should be imported from the top-level `marshmallow` module.

Methods decorated with
`pre_load <marshmallow.decorators.pre_load>`, `post_load <marshmallow.decorators.post_load>`,
`pre_dump <marshmallow.decorators.pre_dump>`, `post_dump <marshmallow.decorators.post_dump>`,
and `validates_schema <marshmallow.decorators.validates_schema>` receive
``many`` as a keyword argument. In addition, `pre_load <marshmallow.decorators.pre_load>`,
`post_load <marshmallow.decorators.post_load>`,
and `validates_schema <marshmallow.decorators.validates_schema>` receive
``partial``. If you don't need these arguments, add ``**kwargs`` to your method
signature.


Example: ::

    from marshmallow import (
        Schema,
        pre_load,
        pre_dump,
        post_load,
        validates_schema,
        validates,
        fields,
        ValidationError,
    )


    class UserSchema(Schema):
        email = fields.Str(required=True)
        age = fields.Integer(required=True)

        @post_load
        def lowerstrip_email(self, item, many, **kwargs):
            item["email"] = item["email"].lower().strip()
            return item

        @pre_load(pass_many=True)
        def remove_envelope(self, data, many, **kwargs):
            namespace = "results" if many else "result"
            return data[namespace]

        @post_dump(pass_many=True)
        def add_envelope(self, data, many, **kwargs):
            namespace = "results" if many else "result"
            return {namespace: data}

        @validates_schema
        def validate_email(self, data, **kwargs):
            if len(data["email"]) < 3:
                raise ValidationError("Email must be more than 3 characters", "email")

        @validates("age")
        def validate_age(self, data, **kwargs):
            if data < 14:
                raise ValidationError("Too young!")

.. note::
    These decorators only work with instance methods. Class and static
    methods are not supported.

.. warning::
    The invocation order of decorated methods of the same type is not guaranteed.
    If you need to guarantee order of different processing steps, you should put
    them in the same processing method.
"""

from __future__ import annotations

import functools
from collections import defaultdict
from typing import Any, Callable, cast

PRE_DUMP = "pre_dump"
POST_DUMP = "post_dump"
PRE_LOAD = "pre_load"
POST_LOAD = "post_load"
VALIDATES = "validates"
VALIDATES_SCHEMA = "validates_schema"


class MarshmallowHook:
    __marshmallow_hook__: dict[str, list[tuple[bool, Any]]] | None = None


def validates(field_name: str) -> Callable[..., Any]:
    """Register a field validator.

    :param field_name: Name of the field that the method validates.
    """
    return set_hook(None, VALIDATES, field_name=field_name)


def validates_schema(
    fn: Callable[..., Any] | None = None,
    pass_many: bool = False,  # noqa: FBT001, FBT002
    pass_original: bool = False,  # noqa: FBT001, FBT002
    skip_on_field_errors: bool = True,  # noqa: FBT001, FBT002
) -> Callable[..., Any]:
    """Register a schema-level validator.

    By default it receives a single object at a time, transparently handling the ``many``
    argument passed to the `Schema <marshmallow.Schema>`'s :func:`~marshmallow.Schema.validate` call.
    If ``pass_many=True``, the raw data (which may be a collection) is passed.

    If ``pass_original=True``, the original data (before unmarshalling) will be passed as
    an additional argument to the method.

    If ``skip_on_field_errors=True``, this validation method will be skipped whenever
    validation errors have been detected when validating fields.

    .. versionchanged:: 3.0.0b1
        ``skip_on_field_errors`` defaults to `True`.

    .. versionchanged:: 3.0.0
        ``partial`` and ``many`` are always passed as keyword arguments to
        the decorated method.
    """
    return set_hook(
        fn,
        VALIDATES_SCHEMA,
        many=pass_many,
        pass_original=pass_original,
        skip_on_field_errors=skip_on_field_errors,
    )


def pre_dump(
    fn: Callable[..., Any] | None = None,
    pass_many: bool = False,  # noqa: FBT001, FBT002
) -> Callable[..., Any]:
    """Register a method to invoke before serializing an object. The method
    receives the object to be serialized and returns the processed object.

    By default it receives a single object at a time, transparently handling the ``many``
    argument passed to the `Schema <marshmallow.Schema>`'s :func:`~marshmallow.Schema.dump` call.
    If ``pass_many=True``, the raw data (which may be a collection) is passed.

    .. versionchanged:: 3.0.0
        ``many`` is always passed as a keyword arguments to the decorated method.
    """
    return set_hook(fn, PRE_DUMP, many=pass_many)


def post_dump(
    fn: Callable[..., Any] | None = None,
    pass_many: bool = False,  # noqa: FBT001, FBT002
    pass_original: bool = False,  # noqa: FBT001, FBT002
) -> Callable[..., Any]:
    """Register a method to invoke after serializing an object. The method
    receives the serialized object and returns the processed object.

    By default it receives a single object at a time, transparently handling the ``many``
    argument passed to the `Schema <marshmallow.Schema>`'s :func:`~marshmallow.Schema.dump` call.
    If ``pass_many=True``, the raw data (which may be a collection) is passed.

    If ``pass_original=True``, the original data (before serializing) will be passed as
    an additional argument to the method.

    .. versionchanged:: 3.0.0
        ``many`` is always passed as a keyword arguments to the decorated method.
    """
    return set_hook(fn, POST_DUMP, many=pass_many, pass_original=pass_original)


def pre_load(
    fn: Callable[..., Any] | None = None,
    pass_many: bool = False,  # noqa: FBT001, FBT002
) -> Callable[..., Any]:
    """Register a method to invoke before deserializing an object. The method
    receives the data to be deserialized and returns the processed data.

    By default it receives a single object at a time, transparently handling the ``many``
    argument passed to the `Schema <marshmallow.Schema>`'s :func:`~marshmallow.Schema.load` call.
    If ``pass_many=True``, the raw data (which may be a collection) is passed.

    .. versionchanged:: 3.0.0
        ``partial`` and ``many`` are always passed as keyword arguments to
        the decorated method.
    """
    return set_hook(fn, PRE_LOAD, many=pass_many)


def post_load(
    fn: Callable[..., Any] | None = None,
    pass_many: bool = False,  # noqa: FBT001, FBT002
    pass_original: bool = False,  # noqa: FBT001, FBT002
) -> Callable[..., Any]:
    """Register a method to invoke after deserializing an object. The method
    receives the deserialized data and returns the processed data.

    By default it receives a single object at a time, transparently handling the ``many``
    argument passed to the `Schema <marshmallow.Schema>`'s :func:`~marshmallow.Schema.load` call.
    If ``pass_many=True``, the raw data (which may be a collection) is passed.

    If ``pass_original=True``, the original data (before deserializing) will be passed as
    an additional argument to the method.

    .. versionchanged:: 3.0.0
        ``partial`` and ``many`` are always passed as keyword arguments to
        the decorated method.
    """
    return set_hook(fn, POST_LOAD, many=pass_many, pass_original=pass_original)


def set_hook(
    fn: Callable[..., Any] | None,
    tag: str,
    many: bool = False,  # noqa: FBT001, FBT002
    **kwargs: Any,
) -> Callable[..., Any]:
    """Mark decorated function as a hook to be picked up later.
    You should not need to use this method directly.

    .. note::
        Currently only works with functions and instance methods. Class and
        static methods are not supported.

    :return: Decorated function if supplied, else this decorator with its args
        bound.
    """
    # Allow using this as either a decorator or a decorator factory.
    if fn is None:
        return functools.partial(set_hook, tag=tag, many=many, **kwargs)

    # Set a __marshmallow_hook__ attribute instead of wrapping in some class,
    # because I still want this to end up as a normal (unbound) method.
    function = cast(MarshmallowHook, fn)
    try:
        hook_config = function.__marshmallow_hook__
    except AttributeError:
        function.__marshmallow_hook__ = hook_config = defaultdict(list)
    # Also save the kwargs for the tagged function on
    # __marshmallow_hook__, keyed by <tag>
    if hook_config is not None:
        hook_config[tag].append((many, kwargs))

    return fn



"""Hypothetical Document Embeddings.

https://arxiv.org/abs/2212.10496
"""
from __future__ import annotations

from typing import Any, Dict, List, Optional

import numpy as np
from langchain_core.callbacks import CallbackManagerForChainRun
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import BasePromptTemplate
from langchain_core.pydantic_v1 import Extra

from langchain.chains.base import Chain
from langchain.chains.hyde.prompts import PROMPT_MAP
from langchain.chains.llm import LLMChain


class HypotheticalDocumentEmbedder(Chain, Embeddings):
    """Generate hypothetical document for query, and then embed that.

    Based on https://arxiv.org/abs/2212.10496
    """

    base_embeddings: Embeddings
    llm_chain: LLMChain

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    @property
    def input_keys(self) -> List[str]:
        """Input keys for Hyde's LLM chain."""
        return self.llm_chain.input_keys

    @property
    def output_keys(self) -> List[str]:
        """Output keys for Hyde's LLM chain."""
        return self.llm_chain.output_keys

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Call the base embeddings."""
        return self.base_embeddings.embed_documents(texts)

    def combine_embeddings(self, embeddings: List[List[float]]) -> List[float]:
        """Combine embeddings into final embeddings."""
        return list(np.array(embeddings).mean(axis=0))

    def embed_query(self, text: str) -> List[float]:
        """Generate a hypothetical document and embedded it."""
        var_name = self.llm_chain.input_keys[0]
        result = self.llm_chain.generate([{var_name: text}])
        documents = [generation.text for generation in result.generations[0]]
        embeddings = self.embed_documents(documents)
        return self.combine_embeddings(embeddings)

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, str]:
        """Call the internal llm chain."""
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        return self.llm_chain(inputs, callbacks=_run_manager.get_child())

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        base_embeddings: Embeddings,
        prompt_key: Optional[str] = None,
        custom_prompt: Optional[BasePromptTemplate] = None,
        **kwargs: Any,
    ) -> HypotheticalDocumentEmbedder:
        """Load and use LLMChain with either a specific prompt key or custom prompt."""
        if custom_prompt is not None:
            prompt = custom_prompt
        elif prompt_key is not None and prompt_key in PROMPT_MAP:
            prompt = PROMPT_MAP[prompt_key]
        else:
            raise ValueError(
                f"Must specify prompt_key if custom_prompt not provided. Should be one "
                f"of {list(PROMPT_MAP.keys())}."
            )

        llm_chain = LLMChain(llm=llm, prompt=prompt)
        return cls(base_embeddings=base_embeddings, llm_chain=llm_chain, **kwargs)

    @property
    def _chain_type(self) -> str:
        return "hyde_chain"



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Encoding and validation utils for the messaging (FCM) module."""

import datetime
import json
import math
import numbers
import re

import firebase_admin._messaging_utils as _messaging_utils


class Message:
    """A message that can be sent via Firebase Cloud Messaging.

    Contains payload information as well as recipient information. In particular, the message must
    contain exactly one of token, topic or condition fields.

    Args:
        data: A dictionary of data fields (optional). All keys and values in the dictionary must be
            strings.
        notification: An instance of ``messaging.Notification`` (optional).
        android: An instance of ``messaging.AndroidConfig`` (optional).
        webpush: An instance of ``messaging.WebpushConfig`` (optional).
        apns: An instance of ``messaging.ApnsConfig`` (optional).
        fcm_options: An instance of ``messaging.FCMOptions`` (optional).
        token: The registration token of the device to which the message should be sent (optional).
        topic: Name of the FCM topic to which the message should be sent (optional). Topic name
            may contain the ``/topics/`` prefix.
        condition: The FCM condition to which the message should be sent (optional).
    """

    def __init__(self, data=None, notification=None, android=None, webpush=None, apns=None,
                 fcm_options=None, token=None, topic=None, condition=None):
        self.data = data
        self.notification = notification
        self.android = android
        self.webpush = webpush
        self.apns = apns
        self.fcm_options = fcm_options
        self.token = token
        self.topic = topic
        self.condition = condition

    def __str__(self):
        return json.dumps(self, cls=MessageEncoder, sort_keys=True)


class MulticastMessage:
    """A message that can be sent to multiple tokens via Firebase Cloud Messaging.

    Args:
        tokens: A list of registration tokens of targeted devices.
        data: A dictionary of data fields (optional). All keys and values in the dictionary must be
            strings.
        notification: An instance of ``messaging.Notification`` (optional).
        android: An instance of ``messaging.AndroidConfig`` (optional).
        webpush: An instance of ``messaging.WebpushConfig`` (optional).
        apns: An instance of ``messaging.ApnsConfig`` (optional).
        fcm_options: An instance of ``messaging.FCMOptions`` (optional).
    """
    def __init__(self, tokens, data=None, notification=None, android=None, webpush=None, apns=None,
                 fcm_options=None):
        _Validators.check_string_list('MulticastMessage.tokens', tokens)
        if len(tokens) > 500:
            raise ValueError('MulticastMessage.tokens must not contain more than 500 tokens.')
        self.tokens = tokens
        self.data = data
        self.notification = notification
        self.android = android
        self.webpush = webpush
        self.apns = apns
        self.fcm_options = fcm_options


class _Validators:
    """A collection of data validation utilities.

    Methods provided in this class raise ``ValueErrors`` if any validations fail.
    """

    @classmethod
    def check_string(cls, label, value, non_empty=False):
        """Checks if the given value is a string."""
        if value is None:
            return None
        if not isinstance(value, str):
            if non_empty:
                raise ValueError('{0} must be a non-empty string.'.format(label))
            raise ValueError('{0} must be a string.'.format(label))
        if non_empty and not value:
            raise ValueError('{0} must be a non-empty string.'.format(label))
        return value

    @classmethod
    def check_number(cls, label, value):
        if value is None:
            return None
        if not isinstance(value, numbers.Number):
            raise ValueError('{0} must be a number.'.format(label))
        return value

    @classmethod
    def check_string_dict(cls, label, value):
        """Checks if the given value is a dictionary comprised only of string keys and values."""
        if value is None or value == {}:
            return None
        if not isinstance(value, dict):
            raise ValueError('{0} must be a dictionary.'.format(label))
        non_str = [k for k in value if not isinstance(k, str)]
        if non_str:
            raise ValueError('{0} must not contain non-string keys.'.format(label))
        non_str = [v for v in value.values() if not isinstance(v, str)]
        if non_str:
            raise ValueError('{0} must not contain non-string values.'.format(label))
        return value

    @classmethod
    def check_string_list(cls, label, value):
        """Checks if the given value is a list comprised only of strings."""
        if value is None or value == []:
            return None
        if not isinstance(value, list):
            raise ValueError('{0} must be a list of strings.'.format(label))
        non_str = [k for k in value if not isinstance(k, str)]
        if non_str:
            raise ValueError('{0} must not contain non-string values.'.format(label))
        return value

    @classmethod
    def check_number_list(cls, label, value):
        """Checks if the given value is a list comprised only of numbers."""
        if value is None or value == []:
            return None
        if not isinstance(value, list):
            raise ValueError('{0} must be a list of numbers.'.format(label))
        non_number = [k for k in value if not isinstance(k, numbers.Number)]
        if non_number:
            raise ValueError('{0} must not contain non-number values.'.format(label))
        return value

    @classmethod
    def check_analytics_label(cls, label, value):
        """Checks if the given value is a valid analytics label."""
        value = _Validators.check_string(label, value)
        if value is not None and not re.match(r'^[a-zA-Z0-9-_.~%]{1,50}$', value):
            raise ValueError('Malformed {}.'.format(label))
        return value

    @classmethod
    def check_boolean(cls, label, value):
        """Checks if the given value is boolean."""
        if value is None:
            return None
        if not isinstance(value, bool):
            raise ValueError('{0} must be a boolean.'.format(label))
        return value

    @classmethod
    def check_datetime(cls, label, value):
        """Checks if the given value is a datetime."""
        if value is None:
            return None
        if not isinstance(value, datetime.datetime):
            raise ValueError('{0} must be a datetime.'.format(label))
        return value


class MessageEncoder(json.JSONEncoder):
    """A custom ``JSONEncoder`` implementation for serializing Message instances into JSON."""

    @classmethod
    def remove_null_values(cls, dict_value):
        return {k: v for k, v in dict_value.items() if v not in [None, [], {}]}

    @classmethod
    def encode_android(cls, android):
        """Encodes an ``AndroidConfig`` instance into JSON."""
        if android is None:
            return None
        if not isinstance(android, _messaging_utils.AndroidConfig):
            raise ValueError('Message.android must be an instance of AndroidConfig class.')
        result = {
            'collapse_key': _Validators.check_string(
                'AndroidConfig.collapse_key', android.collapse_key),
            'data': _Validators.check_string_dict(
                'AndroidConfig.data', android.data),
            'notification': cls.encode_android_notification(android.notification),
            'priority': _Validators.check_string(
                'AndroidConfig.priority', android.priority, non_empty=True),
            'restricted_package_name': _Validators.check_string(
                'AndroidConfig.restricted_package_name', android.restricted_package_name),
            'ttl': cls.encode_ttl(android.ttl),
            'fcm_options': cls.encode_android_fcm_options(android.fcm_options),
            'direct_boot_ok': _Validators.check_boolean(
                'AndroidConfig.direct_boot_ok', android.direct_boot_ok),
        }
        result = cls.remove_null_values(result)
        priority = result.get('priority')
        if priority and priority not in ('high', 'normal'):
            raise ValueError('AndroidConfig.priority must be "high" or "normal".')
        return result

    @classmethod
    def encode_android_fcm_options(cls, fcm_options):
        """Encodes an ``AndroidFCMOptions`` instance into JSON."""
        if fcm_options is None:
            return None
        if not isinstance(fcm_options, _messaging_utils.AndroidFCMOptions):
            raise ValueError('AndroidConfig.fcm_options must be an instance of '
                             'AndroidFCMOptions class.')
        result = {
            'analytics_label': _Validators.check_analytics_label(
                'AndroidFCMOptions.analytics_label', fcm_options.analytics_label),
        }
        result = cls.remove_null_values(result)
        return result

    @classmethod
    def encode_ttl(cls, ttl):
        """Encodes an ``AndroidConfig`` ``TTL`` duration into a string."""
        if ttl is None:
            return None
        if isinstance(ttl, numbers.Number):
            ttl = datetime.timedelta(seconds=ttl)
        if not isinstance(ttl, datetime.timedelta):
            raise ValueError('AndroidConfig.ttl must be a duration in seconds or an instance of '
                             'datetime.timedelta.')
        total_seconds = ttl.total_seconds()
        if total_seconds < 0:
            raise ValueError('AndroidConfig.ttl must not be negative.')
        seconds = int(math.floor(total_seconds))
        nanos = int((total_seconds - seconds) * 1e9)
        if nanos:
            return '{0}.{1}s'.format(seconds, str(nanos).zfill(9))
        return '{0}s'.format(seconds)

    @classmethod
    def encode_milliseconds(cls, label, msec):
        """Encodes a duration in milliseconds into a string."""
        if msec is None:
            return None
        if isinstance(msec, numbers.Number):
            msec = datetime.timedelta(milliseconds=msec)
        if not isinstance(msec, datetime.timedelta):
            raise ValueError('{0} must be a duration in milliseconds or an instance of '
                             'datetime.timedelta.'.format(label))
        total_seconds = msec.total_seconds()
        if total_seconds < 0:
            raise ValueError('{0} must not be negative.'.format(label))
        seconds = int(math.floor(total_seconds))
        nanos = int((total_seconds - seconds) * 1e9)
        if nanos:
            return '{0}.{1}s'.format(seconds, str(nanos).zfill(9))
        return '{0}s'.format(seconds)

    @classmethod
    def encode_android_notification(cls, notification):
        """Encodes an ``AndroidNotification`` instance into JSON."""
        if notification is None:
            return None
        if not isinstance(notification, _messaging_utils.AndroidNotification):
            raise ValueError('AndroidConfig.notification must be an instance of '
                             'AndroidNotification class.')
        result = {
            'body': _Validators.check_string(
                'AndroidNotification.body', notification.body),
            'body_loc_args': _Validators.check_string_list(
                'AndroidNotification.body_loc_args', notification.body_loc_args),
            'body_loc_key': _Validators.check_string(
                'AndroidNotification.body_loc_key', notification.body_loc_key),
            'click_action': _Validators.check_string(
                'AndroidNotification.click_action', notification.click_action),
            'color': _Validators.check_string(
                'AndroidNotification.color', notification.color, non_empty=True),
            'icon': _Validators.check_string(
                'AndroidNotification.icon', notification.icon),
            'sound': _Validators.check_string(
                'AndroidNotification.sound', notification.sound),
            'tag': _Validators.check_string(
                'AndroidNotification.tag', notification.tag),
            'title': _Validators.check_string(
                'AndroidNotification.title', notification.title),
            'title_loc_args': _Validators.check_string_list(
                'AndroidNotification.title_loc_args', notification.title_loc_args),
            'title_loc_key': _Validators.check_string(
                'AndroidNotification.title_loc_key', notification.title_loc_key),
            'channel_id': _Validators.check_string(
                'AndroidNotification.channel_id', notification.channel_id),
            'image': _Validators.check_string(
                'image', notification.image),
            'ticker': _Validators.check_string(
                'AndroidNotification.ticker', notification.ticker),
            'sticky': notification.sticky,
            'event_time': _Validators.check_datetime(
                'AndroidNotification.event_timestamp', notification.event_timestamp),
            'local_only': notification.local_only,
            'notification_priority': _Validators.check_string(
                'AndroidNotification.priority', notification.priority, non_empty=True),
            'vibrate_timings': _Validators.check_number_list(
                'AndroidNotification.vibrate_timings_millis', notification.vibrate_timings_millis),
            'default_vibrate_timings': notification.default_vibrate_timings,
            'default_sound': notification.default_sound,
            'default_light_settings': notification.default_light_settings,
            'light_settings': cls.encode_light_settings(notification.light_settings),
            'visibility': _Validators.check_string(
                'AndroidNotification.visibility', notification.visibility, non_empty=True),
            'notification_count': _Validators.check_number(
                'AndroidNotification.notification_count', notification.notification_count)
        }
        result = cls.remove_null_values(result)
        color = result.get('color')
        if color and not re.match(r'^#[0-9a-fA-F]{6}$', color):
            raise ValueError(
                'AndroidNotification.color must be in the form #RRGGBB.')
        if result.get('body_loc_args') and not result.get('body_loc_key'):
            raise ValueError(
                'AndroidNotification.body_loc_key is required when specifying body_loc_args.')
        if result.get('title_loc_args') and not result.get('title_loc_key'):
            raise ValueError(
                'AndroidNotification.title_loc_key is required when specifying title_loc_args.')

        event_time = result.get('event_time')
        if event_time:
            # if the datetime instance is not naive (tzinfo is present), convert to UTC
            # otherwise (tzinfo is None) assume the datetime instance is already in UTC
            if event_time.tzinfo is not None:
                event_time = event_time.astimezone(datetime.timezone.utc)
            result['event_time'] = event_time.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

        priority = result.get('notification_priority')
        if priority:
            if priority not in ('min', 'low', 'default', 'high', 'max'):
                raise ValueError('AndroidNotification.priority must be "default", "min", "low", '
                                 '"high" or "max".')
            result['notification_priority'] = 'PRIORITY_' + priority.upper()

        visibility = result.get('visibility')
        if visibility:
            if visibility not in ('private', 'public', 'secret'):
                raise ValueError(
                    'AndroidNotification.visibility must be "private", "public" or "secret".')
            result['visibility'] = visibility.upper()

        vibrate_timings_millis = result.get('vibrate_timings')
        if vibrate_timings_millis:
            vibrate_timing_strings = []
            for msec in vibrate_timings_millis:
                formated_string = cls.encode_milliseconds(
                    'AndroidNotification.vibrate_timings_millis', msec)
                vibrate_timing_strings.append(formated_string)
            result['vibrate_timings'] = vibrate_timing_strings
        return result

    @classmethod
    def encode_light_settings(cls, light_settings):
        """Encodes a ``LightSettings`` instance into JSON."""
        if light_settings is None:
            return None
        if not isinstance(light_settings, _messaging_utils.LightSettings):
            raise ValueError(
                'AndroidNotification.light_settings must be an instance of LightSettings class.')
        result = {
            'color': _Validators.check_string(
                'LightSettings.color', light_settings.color, non_empty=True),
            'light_on_duration': cls.encode_milliseconds(
                'LightSettings.light_on_duration_millis', light_settings.light_on_duration_millis),
            'light_off_duration': cls.encode_milliseconds(
                'LightSettings.light_off_duration_millis',
                light_settings.light_off_duration_millis),
        }
        result = cls.remove_null_values(result)
        light_on_duration = result.get('light_on_duration')
        if not light_on_duration:
            raise ValueError(
                'LightSettings.light_on_duration_millis is required.')

        light_off_duration = result.get('light_off_duration')
        if not light_off_duration:
            raise ValueError(
                'LightSettings.light_off_duration_millis is required.')

        color = result.get('color')
        if not color:
            raise ValueError('LightSettings.color is required.')
        if not re.match(r'^#[0-9a-fA-F]{6}$', color) and not re.match(r'^#[0-9a-fA-F]{8}$', color):
            raise ValueError(
                'LightSettings.color must be in the form #RRGGBB or #RRGGBBAA.')
        if len(color) == 7:
            color = (color+'FF')
        rgba = [int(color[i:i + 2], 16) / 255.0 for i in (1, 3, 5, 7)]
        result['color'] = {'red': rgba[0], 'green': rgba[1],
                           'blue': rgba[2], 'alpha': rgba[3]}
        return result

    @classmethod
    def encode_webpush(cls, webpush):
        """Encodes a ``WebpushConfig`` instance into JSON."""
        if webpush is None:
            return None
        if not isinstance(webpush, _messaging_utils.WebpushConfig):
            raise ValueError('Message.webpush must be an instance of WebpushConfig class.')
        result = {
            'data': _Validators.check_string_dict(
                'WebpushConfig.data', webpush.data),
            'headers': _Validators.check_string_dict(
                'WebpushConfig.headers', webpush.headers),
            'notification': cls.encode_webpush_notification(webpush.notification),
            'fcm_options': cls.encode_webpush_fcm_options(webpush.fcm_options),
        }
        return cls.remove_null_values(result)

    @classmethod
    def encode_webpush_notification(cls, notification):
        """Encodes a ``WebpushNotification`` instance into JSON."""
        if notification is None:
            return None
        if not isinstance(notification, _messaging_utils.WebpushNotification):
            raise ValueError('WebpushConfig.notification must be an instance of '
                             'WebpushNotification class.')
        result = {
            'actions': cls.encode_webpush_notification_actions(notification.actions),
            'badge': _Validators.check_string(
                'WebpushNotification.badge', notification.badge),
            'body': _Validators.check_string(
                'WebpushNotification.body', notification.body),
            'data': notification.data,
            'dir': _Validators.check_string(
                'WebpushNotification.direction', notification.direction),
            'icon': _Validators.check_string(
                'WebpushNotification.icon', notification.icon),
            'image': _Validators.check_string(
                'WebpushNotification.image', notification.image),
            'lang': _Validators.check_string(
                'WebpushNotification.language', notification.language),
            'renotify': notification.renotify,
            'requireInteraction': notification.require_interaction,
            'silent': notification.silent,
            'tag': _Validators.check_string(
                'WebpushNotification.tag', notification.tag),
            'timestamp': _Validators.check_number(
                'WebpushNotification.timestamp_millis', notification.timestamp_millis),
            'title': _Validators.check_string(
                'WebpushNotification.title', notification.title),
            'vibrate': notification.vibrate,
        }
        direction = result.get('dir')
        if direction and direction not in ('auto', 'ltr', 'rtl'):
            raise ValueError('WebpushNotification.direction must be "auto", "ltr" or "rtl".')
        if notification.custom_data is not None:
            if not isinstance(notification.custom_data, dict):
                raise ValueError('WebpushNotification.custom_data must be a dict.')
            for key, value in notification.custom_data.items():
                if key in result:
                    raise ValueError(
                        'Multiple specifications for {0} in WebpushNotification.'.format(key))
                result[key] = value
        return cls.remove_null_values(result)

    @classmethod
    def encode_webpush_notification_actions(cls, actions):
        """Encodes a list of ``WebpushNotificationActions`` into JSON."""
        if actions is None:
            return None
        if not isinstance(actions, list):
            raise ValueError('WebpushConfig.notification.actions must be a list of '
                             'WebpushNotificationAction instances.')
        results = []
        for action in actions:
            if not isinstance(action, _messaging_utils.WebpushNotificationAction):
                raise ValueError('WebpushConfig.notification.actions must be a list of '
                                 'WebpushNotificationAction instances.')
            result = {
                'action': _Validators.check_string(
                    'WebpushNotificationAction.action', action.action),
                'title': _Validators.check_string(
                    'WebpushNotificationAction.title', action.title),
                'icon': _Validators.check_string(
                    'WebpushNotificationAction.icon', action.icon),
            }
            results.append(cls.remove_null_values(result))
        return results

    @classmethod
    def encode_webpush_fcm_options(cls, options):
        """Encodes a ``WebpushFCMOptions`` instance into JSON."""
        if options is None:
            return None
        result = {
            'link': _Validators.check_string('WebpushConfig.fcm_options.link', options.link),
        }
        result = cls.remove_null_values(result)
        link = result.get('link')
        if link is not None and not link.startswith('https://'):
            raise ValueError('WebpushFCMOptions.link must be a HTTPS URL.')
        return result

    @classmethod
    def encode_apns(cls, apns):
        """Encodes an ``APNSConfig`` instance into JSON."""
        if apns is None:
            return None
        if not isinstance(apns, _messaging_utils.APNSConfig):
            raise ValueError('Message.apns must be an instance of APNSConfig class.')
        result = {
            'headers': _Validators.check_string_dict(
                'APNSConfig.headers', apns.headers),
            'payload': cls.encode_apns_payload(apns.payload),
            'fcm_options': cls.encode_apns_fcm_options(apns.fcm_options),
        }
        return cls.remove_null_values(result)

    @classmethod
    def encode_apns_payload(cls, payload):
        """Encodes an ``APNSPayload`` instance into JSON."""
        if payload is None:
            return None
        if not isinstance(payload, _messaging_utils.APNSPayload):
            raise ValueError('APNSConfig.payload must be an instance of APNSPayload class.')
        result = {
            'aps': cls.encode_aps(payload.aps)
        }
        for key, value in payload.custom_data.items():
            result[key] = value
        return cls.remove_null_values(result)

    @classmethod
    def encode_apns_fcm_options(cls, fcm_options):
        """Encodes an ``APNSFCMOptions`` instance into JSON."""
        if fcm_options is None:
            return None
        if not isinstance(fcm_options, _messaging_utils.APNSFCMOptions):
            raise ValueError('APNSConfig.fcm_options must be an instance of APNSFCMOptions class.')
        result = {
            'analytics_label': _Validators.check_analytics_label(
                'APNSFCMOptions.analytics_label', fcm_options.analytics_label),
            'image': _Validators.check_string('APNSFCMOptions.image', fcm_options.image)
        }
        result = cls.remove_null_values(result)
        return result

    @classmethod
    def encode_aps(cls, aps):
        """Encodes an ``Aps`` instance into JSON."""
        if not isinstance(aps, _messaging_utils.Aps):
            raise ValueError('APNSPayload.aps must be an instance of Aps class.')
        result = {
            'alert': cls.encode_aps_alert(aps.alert),
            'badge': _Validators.check_number('Aps.badge', aps.badge),
            'sound': cls.encode_aps_sound(aps.sound),
            'category': _Validators.check_string('Aps.category', aps.category),
            'thread-id': _Validators.check_string('Aps.thread_id', aps.thread_id),
        }
        if aps.content_available is True:
            result['content-available'] = 1
        if aps.mutable_content is True:
            result['mutable-content'] = 1
        if aps.custom_data is not None:
            if not isinstance(aps.custom_data, dict):
                raise ValueError('Aps.custom_data must be a dict.')
            for key, val in aps.custom_data.items():
                _Validators.check_string('Aps.custom_data key', key)
                if key in result:
                    raise ValueError('Multiple specifications for {0} in Aps.'.format(key))
                result[key] = val
        return cls.remove_null_values(result)

    @classmethod
    def encode_aps_sound(cls, sound):
        """Encodes an APNs sound configuration into JSON."""
        if sound is None:
            return None
        if sound and isinstance(sound, str):
            return sound
        if not isinstance(sound, _messaging_utils.CriticalSound):
            raise ValueError(
                'Aps.sound must be a non-empty string or an instance of CriticalSound class.')
        result = {
            'name': _Validators.check_string('CriticalSound.name', sound.name, non_empty=True),
            'volume': _Validators.check_number('CriticalSound.volume', sound.volume),
        }
        if sound.critical:
            result['critical'] = 1
        if not result['name']:
            raise ValueError('CriticalSond.name must be a non-empty string.')
        volume = result['volume']
        if volume is not None and (volume < 0 or volume > 1):
            raise ValueError('CriticalSound.volume must be in the interval [0,1].')
        return cls.remove_null_values(result)

    @classmethod
    def encode_aps_alert(cls, alert):
        """Encodes an ``ApsAlert`` instance into JSON."""
        if alert is None:
            return None
        if isinstance(alert, str):
            return alert
        if not isinstance(alert, _messaging_utils.ApsAlert):
            raise ValueError('Aps.alert must be a string or an instance of ApsAlert class.')
        result = {
            'title': _Validators.check_string('ApsAlert.title', alert.title),
            'subtitle': _Validators.check_string('ApsAlert.subtitle', alert.subtitle),
            'body': _Validators.check_string('ApsAlert.body', alert.body),
            'title-loc-key': _Validators.check_string(
                'ApsAlert.title_loc_key', alert.title_loc_key),
            'title-loc-args': _Validators.check_string_list(
                'ApsAlert.title_loc_args', alert.title_loc_args),
            'loc-key': _Validators.check_string(
                'ApsAlert.loc_key', alert.loc_key),
            'loc-args': _Validators.check_string_list(
                'ApsAlert.loc_args', alert.loc_args),
            'action-loc-key': _Validators.check_string(
                'ApsAlert.action_loc_key', alert.action_loc_key),
            'launch-image': _Validators.check_string(
                'ApsAlert.launch_image', alert.launch_image),
        }
        if result.get('loc-args') and not result.get('loc-key'):
            raise ValueError(
                'ApsAlert.loc_key is required when specifying loc_args.')
        if result.get('title-loc-args') and not result.get('title-loc-key'):
            raise ValueError(
                'ApsAlert.title_loc_key is required when specifying title_loc_args.')
        if alert.custom_data is not None:
            if not isinstance(alert.custom_data, dict):
                raise ValueError('ApsAlert.custom_data must be a dict.')
            for key, val in alert.custom_data.items():
                _Validators.check_string('ApsAlert.custom_data key', key)
                # allow specifying key override because Apple could update API so that key
                # could have unexpected value type
                result[key] = val
        return cls.remove_null_values(result)

    @classmethod
    def encode_notification(cls, notification):
        """Encodes a ``Notification`` instance into JSON."""
        if notification is None:
            return None
        if not isinstance(notification, _messaging_utils.Notification):
            raise ValueError('Message.notification must be an instance of Notification class.')
        result = {
            'body': _Validators.check_string('Notification.body', notification.body),
            'title': _Validators.check_string('Notification.title', notification.title),
            'image': _Validators.check_string('Notification.image', notification.image)
        }
        return cls.remove_null_values(result)

    @classmethod
    def sanitize_topic_name(cls, topic):
        """Removes the /topics/ prefix from the topic name, if present."""
        if not topic:
            return None
        prefix = '/topics/'
        if topic.startswith(prefix):
            topic = topic[len(prefix):]
        # Checks for illegal characters and empty string.
        if not re.match(r'^[a-zA-Z0-9-_\.~%]+$', topic):
            raise ValueError('Malformed topic name.')
        return topic

    def default(self, o): # pylint: disable=method-hidden
        if not isinstance(o, Message):
            return json.JSONEncoder.default(self, o)
        result = {
            'android': MessageEncoder.encode_android(o.android),
            'apns': MessageEncoder.encode_apns(o.apns),
            'condition': _Validators.check_string(
                'Message.condition', o.condition, non_empty=True),
            'data': _Validators.check_string_dict('Message.data', o.data),
            'notification': MessageEncoder.encode_notification(o.notification),
            'token': _Validators.check_string('Message.token', o.token, non_empty=True),
            'topic': _Validators.check_string('Message.topic', o.topic, non_empty=True),
            'webpush': MessageEncoder.encode_webpush(o.webpush),
            'fcm_options': MessageEncoder.encode_fcm_options(o.fcm_options),
        }
        result['topic'] = MessageEncoder.sanitize_topic_name(result.get('topic'))
        result = MessageEncoder.remove_null_values(result)
        target_count = sum([t in result for t in ['token', 'topic', 'condition']])
        if target_count != 1:
            raise ValueError('Exactly one of token, topic or condition must be specified.')
        return result

    @classmethod
    def encode_fcm_options(cls, fcm_options):
        """Encodes an ``FCMOptions`` instance into JSON."""
        if fcm_options is None:
            return None
        if not isinstance(fcm_options, _messaging_utils.FCMOptions):
            raise ValueError('Message.fcm_options must be an instance of FCMOptions class.')
        result = {
            'analytics_label': _Validators.check_analytics_label(
                'FCMOptions.analytics_label', fcm_options.analytics_label),
        }
        result = cls.remove_null_values(result)
        return result
